"""
Tests for itwinai.utils module.
"""

from itwinai.utils import flatten_dict, SignatureInspector


def test_flatten_dict():
    """
    Test flatten dict function.
    """
    dict1 = dict(a=1, b=dict(b1=2, b2=3))

    flattened = flatten_dict(dict1)
    assert flattened.get("a") == 1
    assert flattened.get("b.b1") == 2
    assert flattened.get("b.b2") == 3
    assert len(flattened) == 3


def test_signature_inspector():
    """Test SignatureInspector class."""
    def f():
        ...

    inspector = SignatureInspector(f)
    assert not inspector.has_varargs
    assert not inspector.has_kwargs
    assert inspector.required_params == ()
    assert inspector.min_params_num == 0
    assert inspector.max_params_num == 0

    def f(*args):
        ...

    inspector = SignatureInspector(f)
    assert inspector.has_varargs
    assert not inspector.has_kwargs
    assert inspector.required_params == ()
    assert inspector.min_params_num == 0
    assert inspector.max_params_num == SignatureInspector.INFTY

    def f(foo, *args):
        ...

    inspector = SignatureInspector(f)
    assert inspector.has_varargs
    assert not inspector.has_kwargs
    assert inspector.required_params == ("foo",)
    assert inspector.min_params_num == 1
    assert inspector.max_params_num == SignatureInspector.INFTY

    def f(foo, bar=123):
        ...

    inspector = SignatureInspector(f)
    assert not inspector.has_varargs
    assert not inspector.has_kwargs
    assert inspector.required_params == ("foo",)
    assert inspector.min_params_num == 1
    assert inspector.max_params_num == 2

    def f(foo, *args, bar=123):
        ...

    inspector = SignatureInspector(f)
    assert inspector.has_varargs
    assert not inspector.has_kwargs
    assert inspector.required_params == ("foo",)
    assert inspector.min_params_num == 1
    assert inspector.max_params_num == SignatureInspector.INFTY

    def f(*args, **kwargs):
        ...

    inspector = SignatureInspector(f)
    assert inspector.has_varargs
    assert inspector.has_kwargs
    assert inspector.required_params == ()
    assert inspector.min_params_num == 0
    assert inspector.max_params_num == SignatureInspector.INFTY

    def f(foo, /, bar, *arg, **kwargs):
        ...
    inspector = SignatureInspector(f)
    assert inspector.has_varargs
    assert inspector.has_kwargs
    assert inspector.required_params == ("foo", "bar")
    assert inspector.min_params_num == 2
    assert inspector.max_params_num == SignatureInspector.INFTY

    def f(foo, /, bar, *, hello, **kwargs):
        ...
    inspector = SignatureInspector(f)
    assert not inspector.has_varargs
    assert inspector.has_kwargs
    assert inspector.required_params == ("foo", "bar", "hello")
    assert inspector.min_params_num == 3
    assert inspector.max_params_num == SignatureInspector.INFTY



import torch
import numpy as np
import pandas as pd
from operator import add

import model
from torch.utils.data import DataLoader
from engine import evaluate
from initialization import device, beta, criterion, n_avg, pixel_wise_criterion

# pick the season to study among:
# '' (none, i.e. full dataset), 'winter_', 'spring_', 'summer_', 'autumn_'

# choose wether to evaluate train and test data, and/or projections
past_evaluation = False
future_evaluation = True


if past_evaluation:
    for season in ['winter_', 'spring_', 'summer_', 'autumn_']:
        
        # load previously trained model
        cvae_model = model.ConvVAE().to(device)
        cvae_model.load_state_dict(torch.load(f'../outputs/cvae_model_{season}1d.pth'))
        
        # train set and data loader
        train_time = pd.read_csv(f"../input/dates_train_{season}data.csv")
        train_data = np.load(f"../input/preprocessed_1d_train_{season}data_allssp.npy")
        n_train = len(train_data)
        trainset = [ ( torch.from_numpy(np.reshape(train_data[i], (3, 32, 32))), 
                        train_time['0'][i] ) for i in range(n_train) ]
        trainloader = DataLoader(
            trainset, batch_size=1, shuffle=False
        )

        # test set and data loader
        test_time = pd.read_csv(f"../input/dates_test_{season}data.csv")
        test_data = np.load(f"../input/preprocessed_1d_test_{season}data_allssp.npy")
        n_test = len(test_data)
        testset = [ ( torch.from_numpy(np.reshape(test_data[i], (3, 32, 32))), 
                        test_time['0'][i] ) for i in range(n_test) ]
        testloader = DataLoader(
            testset, batch_size=1, shuffle=False
        )
        
        # average over a few iterations 
        # for a better reconstruction estimate
        train_avg_losses, _, tot_train_losses, _ = evaluate(cvae_model, trainloader, 
                                                        trainset, device, 
                                                        criterion, 
                                                        pixel_wise_criterion)
        test_avg_losses, _, tot_test_losses, _ = evaluate(cvae_model, testloader, 
                                                    testset, device, criterion, 
                                                    pixel_wise_criterion)
        for i in range(1, n_avg):
            train_avg_loss, _, train_losses, _ = evaluate(cvae_model, trainloader, 
                                                    trainset, device, criterion, 
                                                    pixel_wise_criterion)
            tot_train_losses = list(map(add, tot_train_losses, train_losses))
            train_avg_losses += train_avg_loss
            test_avg_loss, _, test_losses, _ = evaluate(cvae_model, testloader,
                                                    testset, device, criterion, 
                                                    pixel_wise_criterion)
            tot_test_losses = list(map(add, tot_test_losses, test_losses))
            test_avg_losses += test_avg_loss
        tot_train_losses = np.array(tot_train_losses)/n_avg
        tot_test_losses = np.array(tot_test_losses)/n_avg
        train_avg_losses = train_avg_losses/n_avg
        test_avg_losses = test_avg_losses/n_avg
        
        pd.DataFrame(tot_train_losses).to_csv(f"../outputs/train_losses_{season}1d_allssp.csv")
        pd.DataFrame(tot_test_losses).to_csv(f"../outputs/test_losses_{season}1d_allssp.csv")
        print('Train average loss:', train_avg_losses)
        print('Test average loss:', test_avg_losses)


if future_evaluation:
    for season in ['winter_', 'spring_', 'summer_', 'autumn_']:
        
        # load previously trained model
        cvae_model = model.ConvVAE().to(device)
        cvae_model.load_state_dict(torch.load(f'../outputs/cvae_model_{season}1d.pth'))

        for scenario in ['585', '370', '245', '126']:
            
            # projection set and data loader
            proj_time = pd.read_csv(f"../input/dates_proj_{season}data.csv")
            proj_data = np.load(f"../input/preprocessed_1d_proj{scenario}_{season}data_allssp.npy")
            n_proj = len(proj_data)
            projset = [ ( torch.from_numpy(np.reshape(proj_data[i], (3, 32, 32))), 
                            proj_time['0'][i] ) for i in range(n_proj) ]
            projloader = DataLoader(
                projset, batch_size=1, shuffle=False
            )

            # get the losses for each data set
            # on various experiments to have representative statistics
            proj_avg_losses, _, tot_proj_losses, _ = evaluate(cvae_model, projloader, 
                                                        projset, device, criterion, 
                                                        pixel_wise_criterion)
            
            for i in range(1, n_avg):
                proj_avg_loss, _, proj_losses, _ = evaluate(cvae_model, projloader, 
                                                        projset, device, criterion, 
                                                        pixel_wise_criterion)
                tot_proj_losses = list(map(add, tot_proj_losses, proj_losses))
                proj_avg_losses += proj_avg_loss
            
            tot_proj_losses = np.array(tot_proj_losses)/n_avg
            proj_avg_losses = proj_avg_losses/n_avg

            # save the losses time series
            pd.DataFrame(tot_proj_losses).to_csv(f"../outputs/proj{scenario}_losses_{season}1d_allssp.csv")
            print(f'SSP{scenario} Projection average loss:', proj_avg_losses, 'for', season[:-1])


from typing import Dict, Any, Optional, Union
import logging
from os.path import join, exists


import tensorflow as tf
import tensorflow.keras as keras

from itwinai.tensorflow.distributed import get_strategy
from itwinai.tensorflow.trainer import TensorflowTrainer
from itwinai.components import monitor_exec

from src.utils import get_network_config, load_model
from src.callbacks import ProcessBenchmark
from src.macros import (
    Network,
    Losses,
    RegularizationStrength,
    Activation
)


class CyclonesTrainer(TensorflowTrainer):
    strategy: tf.distribute.Strategy
    num_workers: int

    def __init__(
        self,
        network: Network,
        activation: Activation,
        regularization_strength: RegularizationStrength,
        learning_rate: float,
        loss: Losses,
        epochs: int,
        micro_batch_size: int,
        global_config: Dict[str, Any],
        kernel_size: Optional[int] = None,
        model_backup: Optional[str] = None,
        rnd_seed: Optional[int] = None,
        verbose: Union[str, int] = 'auto'
    ):
        super().__init__(
            epochs=epochs,
            micro_batch_size=micro_batch_size,
            rnd_seed=rnd_seed,
            verbose=verbose
        )
        self.save_parameters(**self.locals2params(locals()))
        self.global_config = global_config
        self.model_backup = model_backup
        self.network = network.value
        self.activation = activation.value
        self.kernel_size = kernel_size
        self.regularization_strength, self.regularizer = (
            regularization_strength.value
        )

        # Loss name and learning rate
        self.loss_name, self.loss = loss.value
        self.learning_rate = learning_rate

        # Parse global config
        self.dynamic_config(self.global_config)

    @monitor_exec
    def execute(self, train_data, validation_data, channels) -> None:
        # train_size and valid_size are the number of unique elements
        # in the dataset, before calling tf.data.Dataset.repeat(num_epochs)
        train_dataset, train_size = train_data
        valid_dataset, valid_size = validation_data

        # Batch and distribute datasets among strategy's replica.
        # Each batch is further split among the workers
        dist_train_dataset = self.strategy.experimental_distribute_dataset(
            train_dataset.batch(
                self.macro_batch_size, drop_remainder=True,
                num_parallel_calls=tf.data.AUTOTUNE
            )
        )
        dist_valid_dataset = self.strategy.experimental_distribute_dataset(
            valid_dataset.batch(
                self.macro_batch_size, drop_remainder=True,
                num_parallel_calls=tf.data.AUTOTUNE
            )
        )

        # Inside the strategy load the model, data generators and train
        with self.strategy.scope():
            if not self.model_backup:
                model = get_network_config(
                    network=self.network,
                    patch_size=self.patch_size,
                    activation=self.activation,
                    regularizer=self.regularizer,
                    kernel_size=self.kernel_size,
                    channels=channels,
                )
                logging.debug("New model created")
            else:
                model = load_model(model_fpath=self.best_model_name)
                logging.debug(
                    f"Model loaded from backup at {self.best_model_name}")

            optimizer = keras.optimizers.Adam(learning_rate=self.learning_rate)
            metrics = [keras.metrics.MeanAbsoluteError(name="mae")]
            model.compile(loss=self.loss_name,
                          optimizer=optimizer, metrics=metrics)
        logging.debug("Model compiled")

        # Print model summary to check if model's architecture is correct
        print(model.summary())

        # Compute the steps per epoch for train and valid
        steps_per_epoch = train_size // self.macro_batch_size
        validation_steps = valid_size // self.macro_batch_size

        print("macro_batch_size: ", self.macro_batch_size, flush=True)

        # Train the model
        model.fit(
            dist_train_dataset,
            validation_data=dist_valid_dataset,
            steps_per_epoch=steps_per_epoch,
            validation_steps=validation_steps,
            epochs=self.epochs,
            callbacks=self.callbacks,
        )
        logging.debug("Model trained")

        # Save the best model
        model.save(self.last_model_name)
        logging.debug("Saved training history")

    def dynamic_config(self, config: Dict) -> None:
        """Parse configuration generated at runtime."""
        self.experiment_dir = config["experiment_dir"]
        self.run_dir = config["run_dir"]
        self.patch_size = config["patch_size"]

        # Paths
        CHECKPOINTS_DIR = join(self.run_dir, "checkpoints")

        # files and csvs definition
        CHECKPOINTS_FILEPATH = join(CHECKPOINTS_DIR, "model_{epoch:02d}.keras")
        LOSS_METRICS_HISTORY_CSV = join(
            self.run_dir, "loss_metrics_history.csv")
        BENCHMARK_HISTORY_CSV = join(self.run_dir, "benchmark_history.csv")

        self.callbacks = [
            keras.callbacks.EarlyStopping(
                monitor="val_loss",
                patience=100,
                min_delta=0.0001,
                restore_best_weights=True,
                verbose=1,
                mode="min",
            ),
            keras.callbacks.CSVLogger(LOSS_METRICS_HISTORY_CSV),
            ProcessBenchmark(BENCHMARK_HISTORY_CSV),
            keras.callbacks.ModelCheckpoint(
                filepath=CHECKPOINTS_FILEPATH,
                save_best_only=True,
                monitor="val_loss",
                mode="min",
                save_weights_only=False,
                verbose=1,
            ),
        ]

        # Check if model backup exists
        if self.model_backup is not None and not exists(self.model_backup):
            raise FileNotFoundError("Model backup file not found")
        if self.model_backup:
            self.best_model_name = join(self.model_backup, "best_model.keras")
        self.last_model_name = join(self.run_dir, "last_model.keras")



"""
Show how to use DDP, Horovod and DeepSpeed strategies interchangeably
with a large neural network trained on Imagenet dataset, showing how
to use checkpoints.
"""
from typing import Optional
import os
import argparse
import sys
from timeit import default_timer as timer
import time

import torch
import torch.nn.functional as F
import torchvision
from torch.utils.data import DataLoader
from torch.utils.data.distributed import DistributedSampler

import deepspeed
import horovod.torch as hvd

from itwinai.torch.distributed import (
    TorchDistributedStrategy,
    TorchDDPStrategy,
    HorovodStrategy,
    DeepSpeedStrategy,
)
from itwinai.parser import ArgumentParser as ItAIArgumentParser
from itwinai.loggers import EpochTimeTracker
from itwinai.torch.reproducibility import (
    seed_worker, set_seed
)

from utils import imagenet_dataset


def parse_params() -> argparse.Namespace:
    """
    Parse CLI args, which can also be loaded from a configuration file
    using the --config flag:

    >>> train.py --strategy ddp --config base-config.yaml --config foo.yaml
    """
    parser = ItAIArgumentParser(description='PyTorch Imagenet Example')

    # Distributed ML strategy
    parser.add_argument(
        "--strategy", "-s", type=str,
        choices=['ddp', 'horovod', 'deepspeed'],
        default='ddp'
    )

    # Data and logging
    parser.add_argument('--data-dir', default='./',
                        help=('location of the training dataset in the local '
                              'filesystem'))
    parser.add_argument('--log-int', type=int, default=10,
                        help='log interval per training')
    parser.add_argument('--verbose',
                        action=argparse.BooleanOptionalAction,
                        help='Print parsed arguments')
    parser.add_argument('--nworker', type=int, default=0,
                        help=('number of workers in DataLoader (default: 0 -'
                              ' only main)'))
    parser.add_argument('--prefetch', type=int, default=2,
                        help='prefetch data in DataLoader (default: 2)')

    # Model
    parser.add_argument('--batch-size', type=int, default=64,
                        help='input batch size for training (default: 64)')
    parser.add_argument('--epochs', type=int, default=10,
                        help='number of epochs to train (default: 10)')
    parser.add_argument('--lr', type=float, default=0.01,
                        help='learning rate (default: 0.01)')
    parser.add_argument('--momentum', type=float, default=0.5,
                        help='momentum in SGD optimizer (default: 0.5)')
    parser.add_argument('--shuff', action='store_true', default=False,
                        help='shuffle dataset (default: False)')

    # Reproducibility
    parser.add_argument('--rnd-seed', type=Optional[int], default=None,
                        help='seed integer for reproducibility (default: 0)')

    # Distributed ML
    parser.add_argument('--backend', type=str, default='nccl',
                        help='backend for parrallelisation (default: nccl)')
    parser.add_argument('--no-cuda', action='store_true', default=False,
                        help='disables GPGPUs')
    parser.add_argument('--local_rank', type=int, default=-1,
                        help='local rank passed from distributed launcher')

    # Horovod
    parser.add_argument('--fp16-allreduce', action='store_true', default=False,
                        help='use fp16 compression during allreduce')
    parser.add_argument('--use-adasum', action='store_true', default=False,
                        help='use adasum algorithm to do reduction')
    parser.add_argument('--gradient-predivide-factor', type=float, default=1.0,
                        help=('apply gradient pre-divide factor in optimizer '
                              '(default: 1.0)'))

    # DeepSpeed
    parser = deepspeed.add_config_arguments(parser)
    args = parser.parse_args()

    if args.verbose:
        args_list = [f"{key}: {val}" for key, val in args.items()]
        print("PARSED ARGS:\n", '\n'.join(args_list))

    return args


def train(
    model, device, train_loader, optimizer, epoch,
    strategy: TorchDistributedStrategy, args
):
    """
    Training function, representing an epoch.
    """
    model.train()
    t_list = []
    loss_acc = 0
    gwsize = strategy.global_world_size()
    if strategy.is_main_worker:
        print("\n")
    for batch_idx, (data, target) in enumerate(train_loader):
        t = timer()
        data, target = data.to(device), target.to(device)
        optimizer.zero_grad()
        output = model(data)
        loss = F.nll_loss(output, target)
        loss.backward()
        optimizer.step()
        if (strategy.is_main_worker and args.log_int > 0
                and batch_idx % args.log_int == 0):
            print(
                f'Train epoch: {epoch} '
                f'[{batch_idx * len(data)}/{len(train_loader.dataset)/gwsize} '
                f'({100.0 * batch_idx / len(train_loader):.0f}%)]\t\t'
                f'Loss: {loss.item():.6f}')
        t_list.append(timer() - t)
        loss_acc += loss.item()
    if strategy.is_main_worker:
        print('TIMER: train time', sum(t_list) / len(t_list), 's')
    return loss_acc


def main():
    # Parse CLI args
    args = parse_params()

    # Instantiate Strategy
    if args.strategy == 'ddp':
        if (not torch.cuda.is_available()
                or not torch.cuda.device_count() > 1):
            raise RuntimeError('Resources unavailable')

        strategy = TorchDDPStrategy(backend=args.backend)
        distribute_kwargs = {}
    elif args.strategy == 'horovod':
        strategy = HorovodStrategy()
        distribute_kwargs = dict(
            compression=(
                hvd.Compression.fp16 if args.fp16_allreduce
                else hvd.Compression.none
            ),
            op=hvd.Adasum if args.use_adasum else hvd.Average,
            gradient_predivide_factor=args.gradient_predivide_factor
        )
    elif args.strategy == 'deepspeed':
        strategy = DeepSpeedStrategy(backend=args.backend)
        distribute_kwargs = dict(
            config_params=dict(train_micro_batch_size_per_gpu=args.batch_size)
        )
    else:
        raise NotImplementedError(
            f"Strategy {args.strategy} is not recognized/implemented.")
    strategy.init()

    # Check resources availability
    use_cuda = not args.no_cuda and torch.cuda.is_available()
    is_distributed = False
    if use_cuda and torch.cuda.device_count() > 0:
        is_distributed = True

    # Limit # of CPU threads to be used per worker
    # torch.set_num_threads(1)

    # Start the timer for profiling
    st = timer()

    # Set random seed for reproducibility
    torch_prng = set_seed(args.rnd_seed, deterministic_cudnn=False)

    # Get job rank info - rank==0 master gpu
    if is_distributed:
        # local world size - per node
        lwsize = strategy.local_world_size()   # local world size - per run
        gwsize = strategy.global_world_size()   # global world size - per run
        grank = strategy.global_rank()     # global rank - assign per run
        lrank = strategy.local_rank()     # local rank - assign per node
    else:
        # Use a single worker (either on GPU or CPU)
        lwsize = 1
        gwsize = 1
        grank = 0
        lrank = 0

    if strategy.is_main_worker:
        print('TIMER: initialise:', timer()-st, 's')
        print('DEBUG: local ranks:', lwsize, '/ global ranks:', gwsize)
        print('DEBUG: sys.version:', sys.version)
        print('DEBUG: args.data_dir:', args.data_dir)
        print('DEBUG: args.log_int:', args.log_int)
        print('DEBUG: args.nworker:', args.nworker)
        print('DEBUG: args.prefetch:', args.prefetch)
        print('DEBUG: args.batch_size:', args.batch_size)
        print('DEBUG: args.epochs:', args.epochs)
        print('DEBUG: args.lr:', args.lr)
        print('DEBUG: args.momentum:', args.momentum)
        print('DEBUG: args.shuff:', args.shuff)
        print('DEBUG: args.rnd_seed:', args.rnd_seed)
        print('DEBUG: args.backend:', args.backend)
        print('DEBUG: args.no_cuda:', args.no_cuda, '\n')

    # Encapsulate the model on the GPU assigned to the current process
    device = torch.device(
        strategy.device() if use_cuda
        else 'cpu')
    if use_cuda:
        torch.cuda.set_device(lrank)

    # Dataset
    train_dataset = imagenet_dataset(args.data_dir)

    if is_distributed:
        # Distributed sampler restricts data loading to a subset of the dataset
        # exclusive to the current process.
        train_sampler = DistributedSampler(
            train_dataset, num_replicas=gwsize, rank=grank,
            shuffle=(args.shuff and args.rnd_seed is None)
        )

        train_loader = DataLoader(
            train_dataset, batch_size=args.batch_size,
            sampler=train_sampler, num_workers=args.nworker, pin_memory=True,
            persistent_workers=(args.nworker > 1),
            prefetch_factor=args.prefetch, generator=torch_prng,
            worker_init_fn=seed_worker
        )
    else:
        train_loader = DataLoader(
            train_dataset, batch_size=args.batch_size, generator=torch_prng,
            worker_init_fn=seed_worker
        )

    # Create CNN model: resnet 50, resnet101, resnet152
    model = torchvision.models.resnet152()

    # Optimizer
    optimizer = torch.optim.SGD(
        model.parameters(), lr=args.lr, momentum=args.momentum)

    if is_distributed:
        distrib_model, optimizer, _ = strategy.distributed(
            model, optimizer, lr_scheduler=None, **distribute_kwargs
        )

    # Start training loop
    if strategy.is_main_worker:
        print('TIMER: broadcast:', timer()-st, 's')
        print('\nDEBUG: start training')
        print('--------------------------------------------------------')
        nnod = os.environ.get('SLURM_NNODES', 'unk')
        s_name = f"{args.strategy}-it"
        epoch_time_tracker = EpochTimeTracker(
            series_name=s_name,
            csv_file=f"epochtime_{s_name}_{nnod}N.csv"
        )

    et = timer()
    start_epoch = 1
    for epoch in range(start_epoch, args.epochs + 1):
        lt = timer()
        if is_distributed:
            # Inform the sampler that a new epoch started: shuffle
            # may be needed
            train_sampler.set_epoch(epoch)

        # Training
        train(
            model=distrib_model,
            device=device,
            train_loader=train_loader,
            optimizer=optimizer,
            epoch=epoch,
            strategy=strategy,
            args=args
        )

        # Save first epoch timer
        if epoch == start_epoch:
            first_ep_t = timer()-lt

        # Final epoch
        if epoch + 1 == args.epochs:
            train_loader.last_epoch = True

        if strategy.is_main_worker:
            print('TIMER: epoch time:', timer()-lt, 's')
            epoch_time_tracker.add_epoch_time(epoch-1, timer()-lt)

    if strategy.is_main_worker:
        print('\n--------------------------------------------------------')
        print('DEBUG: training results:\n')
        print('TIMER: first epoch time:', first_ep_t, ' s')
        print('TIMER: last epoch time:', timer()-lt, ' s')
        print('TIMER: average epoch time:', (timer()-et)/args.epochs, ' s')
        print('TIMER: total epoch time:', timer()-et, ' s')
        if epoch > 1:
            print('TIMER: total epoch-1 time:',
                  timer()-et-first_ep_t, ' s')
            print('TIMER: average epoch-1 time:',
                  (timer()-et-first_ep_t)/(args.epochs-1), ' s')
        if use_cuda:
            print('DEBUG: memory req:',
                  int(torch.cuda.memory_reserved(lrank)/1024/1024), 'MB')
            print('DEBUG: memory summary:\n\n',
                  torch.cuda.memory_summary(0))

        print(f'TIMER: final time: {timer()-st} s\n')

    time.sleep(1)
    print(f"<Global rank: {strategy.global_rank()}> - TRAINING FINISHED")

    # Clean-up
    if is_distributed:
        strategy.clean_up()


if __name__ == "__main__":
    main()
    sys.exit()



import yaml
import pytest

from itwinai.pipeline import Pipeline
from itwinai.parser import ConfigParser
from itwinai.tests import (
    FakeGetterExec, FakeSplitterExec, FakeTrainerExec, FakeSaverExec
)


def test_slice_into_sub_pipelines():
    """Test slicing the pipeline to obtain
    sub-pipelines as Pipeline objects.
    """
    p = Pipeline(['step1', 'step2', 'step3'])
    sub_pipe1, sub_pipe2 = p[:1], p[1:]
    assert isinstance(sub_pipe1, Pipeline)
    assert isinstance(sub_pipe2, Pipeline)
    assert len(sub_pipe1) == 1
    assert sub_pipe1[0] == "step1"
    assert len(sub_pipe2) == 2

    p = Pipeline(dict(step1="step1", step2="step2", step3="step3"))
    sub_pipe1, sub_pipe2 = p[:1], p[1:]
    assert isinstance(sub_pipe1, Pipeline)
    assert isinstance(sub_pipe2, Pipeline)
    assert len(sub_pipe1) == 1
    assert sub_pipe1["step1"] == "step1"
    assert len(sub_pipe2) == 2


def test_serialization_pipe_list():
    """Test dict serialization of pipeline
    defined as list of BaseComponent objects.
    """
    config = yaml.safe_load(pytest.PIPE_LIST_YAML)
    parser = ConfigParser(config=config)
    pipe = parser.parse_pipeline(
        pipeline_nested_key="my-list-pipeline"
    )

    dict_pipe = pipe.to_dict()
    del dict_pipe['init_args']['name']
    dict_pipe = {"my-list-pipeline": dict_pipe}
    assert dict_pipe == config


def test_serialization_pipe_dict():
    """Test dict serialization of pipeline
    defined as dict of BaseComponent objects.
    """
    config = yaml.safe_load(pytest.PIPE_DICT_YAML)
    parser = ConfigParser(config=config)
    pipe = parser.parse_pipeline(
        pipeline_nested_key="my-dict-pipeline"
    )

    dict_pipe = pipe.to_dict()
    del dict_pipe['init_args']['name']
    dict_pipe = {"my-dict-pipeline": dict_pipe}
    assert dict_pipe == config


def test_arguments_mismatch():
    """Test mismatch of arguments passed among components in a pipeline."""
    pipeline = Pipeline([
        FakeGetterExec(data_uri='http://...'),
        FakeSplitterExec(train_prop=.7),
        FakeTrainerExec(lr=1e-3, batch_size=32),
        # Adapter(policy=[f"{Adapter.INPUT_PREFIX}-1"]),
        FakeSaverExec(save_path="my_model.pth")
    ])
    with pytest.raises(TypeError) as exc_info:
        _ = pipeline.execute()
    assert "received too many input arguments" in str(exc_info.value)

    pipeline = Pipeline([
        FakeGetterExec(data_uri='http://...'),
        FakeTrainerExec(lr=1e-3, batch_size=32),
    ])
    with pytest.raises(TypeError) as exc_info:
        _ = pipeline.execute()
    assert "received too few input arguments" in str(exc_info.value)



import keras
import json


def model_to_json(model: keras.Model, filepath: str):
    """Serialize Keras model to JSON file.

    Args:
        model (keras.Model): Keras model.
        filepath (str): JSON file path.
    """
    with open(filepath, "w") as f:
        json.dump(model.to_json(), f)


def model_from_json(filepath: str) -> keras.Model:
    """Deserialize Keras model from JSON file.

    Args:
        filepath (str): JSON file path.

    Returns:
        keras.Model: loaded Keras model.
    """
    with open(filepath, "r") as f:
        config = json.load(f)
        return keras.models.model_from_json(config)



"""Test distributed training strategies."""

import subprocess
import pytest


@pytest.mark.slurm
def test_distributed_decorator(torch_env):
    """Test function decorator. Needs torchrun cmd."""
    cmd = (f"{torch_env}/bin/torchrun "
           " --nnodes=1 --nproc_per_node=2 --rdzv_id=100 "
           "--rdzv_backend=c10d --rdzv_endpoint=localhost:29400 "
           "tests/torch/distribtued_decorator.py")
    subprocess.run(cmd.split(), check=True)


@pytest.mark.skip(reason="TorchTrainer not implemented yet")
@pytest.mark.slurm
def test_distributed_trainer(torch_env):
    """Test vanilla torch distributed trainer. Needs torchrun cmd."""
    cmd = (f"{torch_env}/bin/torchrun "
           "--nnodes=1 --nproc_per_node=2 --rdzv_id=100 "
           "--rdzv_backend=c10d --rdzv_endpoint=localhost:29400 "
           "tests/torch/torch_dist_trainer.py")
    subprocess.run(cmd.split(), check=True)



"""Custom types definition."""

from typing import Callable

import torch
from torch import nn

#: Torch Loss function.
Loss = nn.Module

#: Torch learning rate scheduler
LrScheduler = nn.Module

#: Torch data batch sampled by a ``DataLoader``.
Batch = torch.Tensor

#: Torch metric function provided by ``torchmetrics`` library.
Metric = Callable


class UninitializedStrategyError(Exception):
    """Error raised when a strategy has not been initialized."""


class DistributedStrategyError(Exception):
    """Error raised when a strategy has already been initialized."""



import pytest

from itwinai.components import Trainer, Adapter
from itwinai.pipeline import Pipeline
from itwinai.tests import (
    FakeGetterExec, FakeSplitterExec, FakeTrainerExec, FakeSaverExec
)
from itwinai.serialization import SerializationError


def test_serializable():
    """Test serialization of components."""
    comp = FakeGetterExec(data_uri='http://...')
    dict_serializ = comp.to_dict()
    assert isinstance(dict_serializ, dict)
    assert comp.name == "FakeGetterExec"
    assert dict_serializ == dict(
        class_path="itwinai.tests.dummy_components.FakeGetterExec",
        init_args=dict(data_uri='http://...', name=None)
    )

    # List
    comp = FakeGetterExec(data_uri=[1, 2, 3])
    dict_serializ = comp.to_dict()
    assert isinstance(dict_serializ, dict)
    assert comp.name == "FakeGetterExec"
    assert dict_serializ == dict(
        class_path="itwinai.tests.dummy_components.FakeGetterExec",
        init_args=dict(data_uri=[1, 2, 3], name=None)
    )

    # Tuple
    comp = FakeGetterExec(data_uri=(1, 2, 3))
    dict_serializ = comp.to_dict()
    assert isinstance(dict_serializ, dict)
    assert comp.name == "FakeGetterExec"
    assert dict_serializ == dict(
        class_path="itwinai.tests.dummy_components.FakeGetterExec",
        init_args=dict(data_uri=[1, 2, 3], name=None)
    )

    # Set
    comp = FakeGetterExec(data_uri={1, 2, 3})
    dict_serializ = comp.to_dict()
    assert isinstance(dict_serializ, dict)
    assert comp.name == "FakeGetterExec"
    assert dict_serializ == dict(
        class_path="itwinai.tests.dummy_components.FakeGetterExec",
        init_args=dict(data_uri=[1, 2, 3], name=None)
    )

    # Dict
    comp = FakeGetterExec(data_uri=dict(foo=12, bar="123", hl=3.14))
    dict_serializ = comp.to_dict()
    assert isinstance(dict_serializ, dict)
    assert comp.name == "FakeGetterExec"
    assert dict_serializ == dict(
        class_path="itwinai.tests.dummy_components.FakeGetterExec",
        init_args=dict(data_uri=dict(foo=12, bar="123", hl=3.14), name=None)
    )

    # Non serializable obj
    class NonSerializable:
        ...

    comp = FakeGetterExec(data_uri=NonSerializable())
    with pytest.raises(SerializationError) as exc_info:
        dict_serializ = comp.to_dict()
    assert ("is not a Python built-in type and it does not extend"
            in str(exc_info.value))

    # Local component class
    class MyTrainer(Trainer):
        def execute(self):
            ...

    comp = MyTrainer()
    with pytest.raises(SerializationError) as exc_info:
        dict_serializ = comp.to_dict()
    assert ("is defined locally, which is not supported for serialization."
            in str(exc_info.value))


def test_adapter():
    """Test Adapter component."""
    prefix = Adapter.INPUT_PREFIX
    adapter = Adapter(
        policy=[f"{prefix}{3-i}" for i in range(4)]
    )
    result = adapter.execute(0, 1, 2, 3)
    assert result == (3, 2, 1, 0)

    result = adapter.execute(*tuple(range(100)))
    assert result == (3, 2, 1, 0)

    adapter = Adapter(
        policy=[f"{prefix}0" for i in range(4)]
    )
    result = adapter.execute(0, 1, 2, 3)
    assert result == (0, 0, 0, 0)

    adapter = Adapter(
        policy=[f"{prefix}{i % 2}" for i in range(4)]
    )
    result = adapter.execute(0, 1, 2, 3)
    assert result == (0, 1, 0, 1)

    adapter = Adapter(
        policy=[f"{prefix}2", "hello", "world", 3.14]
    )
    result = adapter.execute(0, 1, 2, 3)
    assert result == (2, "hello", "world", 3.14)

    adapter = Adapter(
        policy=[1, 3, 5, 7, 11]
    )
    result = adapter.execute(0, 1, 2, 3)
    assert result == (1, 3, 5, 7, 11)

    adapter = Adapter(
        policy=[f"{prefix}{9-i}" for i in range(10)]
    )
    with pytest.raises(IndexError) as exc_info:
        result = adapter.execute(0, 1)
    assert str(exc_info.value) == (
        "The args received as input by 'Adapter' are not consistent with "
        "the given adapter policy because input args are too few! Input "
        "args are 2 but the policy foresees at least 10 items."
    )

    adapter = Adapter(
        policy=[]
    )
    result = adapter.execute(*tuple(range(100)))
    assert result == ()


@pytest.mark.integration
def test_adapter_integration_pipeline():
    """Test integration of Adapter component in the pipeline,
    connecting other components.
    """
    pipeline = Pipeline([
        FakeGetterExec(data_uri='http://...'),
        FakeSplitterExec(train_prop=.7),
        FakeTrainerExec(lr=1e-3, batch_size=32),
        Adapter(policy=[f"{Adapter.INPUT_PREFIX}-1"]),
        FakeSaverExec(save_path="my_model.pth")
    ])
    saved_model = pipeline.execute()
    assert saved_model == FakeTrainerExec.model



import tensorflow as tf


def coo_rot180(data):
    X, y = data
    patch_size = X.shape[0]
    X = tf.image.rot90(X, k=2)
    y1 = [-1., -1.]
    if y[0] != -1:
        y1 = [-y[0] + patch_size - 1, -y[1] + patch_size - 1]
    return (X, y1)


def coo_left_right(data):
    X, y = data
    patch_size = X.shape[0]
    X = tf.image.flip_left_right(X)
    y1 = [-1., -1.]
    if y[0] != -1:
        y1 = [y[0], - y[1] + patch_size - 1]
    return (X, y1)


def coo_up_down(data):
    X, y = data
    patch_size = X.shape[0]
    X = tf.image.flip_up_down(X)
    y1 = [-1., -1.]
    if y[0] != -1:
        y1 = [- y[0] + patch_size - 1, y[1]]
    return (X, y1)


def msk_rot180(data):
    X, Y = data
    X = tf.image.rot90(X, k=2)
    Y = tf.image.rot90(Y, k=2)
    return (X, Y)


def msk_left_right(data):
    X, Y = data
    X = tf.image.flip_left_right(X)
    Y = tf.image.flip_left_right(Y)
    return (X, Y)


def msk_up_down(data):
    X, Y = data
    X = tf.image.flip_up_down(X)
    Y = tf.image.flip_up_down(Y)
    return (X, Y)



"""
Utilities for itwinai package.
"""
from typing import Dict, Type, Callable, Tuple, Hashable
import sys
import inspect
from collections.abc import MutableMapping
import yaml


def load_yaml(path: str) -> Dict:
    """Load YAML file as dict.

    Args:
        path (str): path to YAML file.

    Raises:
        yaml.YAMLError: for loading/parsing errors.

    Returns:
        Dict: nested dict representation of parsed YAML file.
    """
    with open(path, "r", encoding="utf-8") as yaml_file:
        try:
            loaded_config = yaml.safe_load(yaml_file)
        except yaml.YAMLError as exc:
            print(exc)
            raise exc
    return loaded_config


def dynamically_import_class(name: str) -> Type:
    """
    Dynamically import class by module path.
    Adapted from https://stackoverflow.com/a/547867

    Args:
        name (str): path to the class (e.g., mypackage.mymodule.MyClass)

    Returns:
        __class__: class type.
    """
    try:
        module, class_name = name.rsplit(".", 1)
        mod = __import__(module, fromlist=[class_name])
        klass = getattr(mod, class_name)
    except ModuleNotFoundError as err:
        print(
            f"Module not found when trying to dynamically import '{name}'. "
            "Make sure that the module's file is reachable from your current "
            "directory."
        )
        raise err
    except Exception as err:
        print(
            f"Exception occurred when trying to dynamically import '{name}'. "
            "Make sure that the module's file is reachable from your current "
            "directory and that the class is present in that module."
        )
        raise err

    return klass


def flatten_dict(
    d: MutableMapping, parent_key: str = "", sep: str = "."
) -> MutableMapping:
    """Flatten dictionary

    Args:
        d (MutableMapping): nested dictionary to flatten
        parent_key (str, optional): prefix for all keys. Defaults to ''.
        sep (str, optional): separator for nested key concatenation.
            Defaults to '.'.

    Returns:
        MutableMapping: flattened dictionary with new keys.
    """
    items = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, MutableMapping):
            items.extend(flatten_dict(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)


class SignatureInspector:
    """Provides the functionalities to inspect the signature of a function
    or a method.

    Args:
        func (Callable): function to be inspected.
    """

    INFTY: int = sys.maxsize

    def __init__(self, func: Callable) -> None:
        self.func = func
        self.func_params = inspect.signature(func).parameters.items()

    @property
    def has_varargs(self) -> bool:
        """Checks if the function has ``*args`` parameter."""
        return any(map(
            lambda p: p[1].kind == p[1].VAR_POSITIONAL,
            self.func_params
        ))

    @property
    def has_kwargs(self) -> bool:
        """Checks if the function has ``**kwargs`` parameter."""
        return any(map(
            lambda p: p[1].kind == p[1].VAR_KEYWORD,
            self.func_params
        ))

    @property
    def required_params(self) -> Tuple[str]:
        """Names of required parameters. Class method's 'self' is skipped."""
        required_params = list(filter(
            lambda p: (p[0] != 'self' and p[1].default == inspect._empty
                       and p[1].kind != p[1].VAR_POSITIONAL
                       and p[1].kind != p[1].VAR_KEYWORD),
            self.func_params
        ))
        return tuple(map(lambda p: p[0], required_params))

    @property
    def min_params_num(self) -> int:
        """Minimum number of arguments required."""
        return len(self.required_params)

    @property
    def max_params_num(self) -> int:
        """Max number of supported input arguments.
        If no limit, ``SignatureInspector.INFTY`` is returned.
        """
        if self.has_kwargs or self.has_varargs:
            return self.INFTY
        return len(self.func_params)


def str_to_slice(interval: str) -> slice:
    """Transform string interval to Python slice.
    Example: "1:17:3" -> slice(1,17,3)

    Args:
        interval (str): interval to parse.

    Raises:
        ValueError: when interval is invalid.

    Returns:
        slice: parsed slice.
    """
    import re
    # TODO: add support for slices starting with empty index
    # e.g., :20:3
    if not re.match(r"\d+(:\d+)?(:\d+)?", interval):
        raise ValueError(
            f"Received invalid interval for slice: '{interval}'"
        )
    if ":" in interval:
        return slice(*map(
            lambda x: int(x.strip()) if x.strip() else None,
            interval.split(':')
        ))
    return int(interval)


def clear_key(
        my_dict: Dict,
        dict_name: str,
        key: Hashable,
        complain: bool = True
) -> Dict:
    """Remove key from dictionary if present and complain.

    Args:
        my_dict (Dict): Dictionary.
        dict_name (str): name of the dictionary.
        key (Hashable): Key to remove.
    """
    if key in my_dict:
        if complain:
            print(
                f"Field '{key}' should not be present "
                f"in dictionary '{dict_name}'"
            )
        del my_dict[key]
    return my_dict



#!/usr/bin/env python
# coding: utf-8

# ## Preprocess Data for seasonal VAEs

# The aim of this notebook is to translate numpy files (.npy) and time series (.csv)
# of daily maximum temperature to four numpy 3D-arrays: one for each season.
# These output arrays can easily be read for training and evaluating the
# Convolutional Variational AutoEncoder model.

# #### 0. Libraries

import numpy as np
import pandas as pd


# #### 1. Load Data to xarray

# choose the needed number of members
n_memb = 1

# define relevant scenarios
#scenarios = ["126", "245", "370", "585"]
scenarios = ["245", "585"]

# Load preprocessed "daily temperature images" and time series

train_images = np.load("../input/preprocessed_2d_train_data_allssp.npy")
test_images = np.load("../input/preprocessed_2d_test_data_allssp.npy")
train_time = pd.read_csv("../input/dates_train_data.csv")
test_time = pd.read_csv("../input/dates_test_data.csv")


# #### 2. Split Yearly Data into Four Seasonal Datasets

# split daily data into seasons


def season_split(
    images: np.ndarray,
    time: pd.DataFrame,
    dataset_type: str,
    n_memb: int,
    scenario: str = "",
) -> tuple[list[np.ndarray], list[pd.DataFrame]]:
    """
    Splits and returns the data sets (climate variable and time) per season.

    Parameters:
    images: temperature data sets
    time: time data sets
    dataset_type: 'train', 'test', 'proj'
    n_memb: number of ensemble members
    scenario: '' for train or test data,
              or '126', '245', '370', '585' for projections
    """

    n_years = int(len(images) / 365)

    # 1st April = index 90
    # 1st July = index 181
    # 1st October = index 273
    winter_index = [365 * i + j for i in range(n_years) for j in range(90)]
    spring_index = [365 * i + j for i in range(n_years) for j in range(90, 181)]
    summer_index = [365 * i + j for i in range(n_years) for j in range(181, 273)]
    autumn_index = [365 * i + j for i in range(n_years) for j in range(273, 365)]

    winter_images = images[winter_index]
    spring_images = images[spring_index]
    summer_images = images[summer_index]
    autumn_images = images[autumn_index]

    winter_time = time.loc[winter_index].iloc[:, 1]
    spring_time = time.loc[spring_index].iloc[:, 1]
    summer_time = time.loc[summer_index].iloc[:, 1]
    autumn_time = time.loc[autumn_index].iloc[:, 1]

    # save results as an input for CVAE training
    np.save(
        f"../input/preprocessed_1d_{dataset_type}{scenario}_winter_data_{n_memb}memb.npy",
        winter_images,
    )
    np.save(
        f"../input/preprocessed_1d_{dataset_type}{scenario}_spring_data_{n_memb}memb.npy",
        spring_images,
    )
    np.save(
        f"../input/preprocessed_1d_{dataset_type}{scenario}_summer_data_{n_memb}memb.npy",
        summer_images,
    )
    np.save(
        f"../input/preprocessed_1d_{dataset_type}{scenario}_autumn_data_{n_memb}memb.npy",
        autumn_images,
    )
    pd.DataFrame(winter_time).to_csv(
        f"../input/dates_{dataset_type}_winter_data_{n_memb}memb.csv"
    )
    pd.DataFrame(spring_time).to_csv(
        f"../input/dates_{dataset_type}_spring_data_{n_memb}memb.csv"
    )
    pd.DataFrame(summer_time).to_csv(
        f"../input/dates_{dataset_type}_summer_data_{n_memb}memb.csv"
    )
    pd.DataFrame(autumn_time).to_csv(
        f"../input/dates_{dataset_type}_autumn_data_{n_memb}memb.csv"
    )

    season_images = [winter_images, spring_images, summer_images, autumn_images]
    season_time = winter_time, spring_time, summer_time, autumn_time

    return season_images, season_time


# #### 3. Apply to Train and Test Datasets

train_season_images, train_season_time = season_split(
    train_images, train_time, "train", n_memb
)


test_season_images, test_season_time = season_split(
    test_images, test_time, "test", n_memb
)


# #### 4. Apply to Projection Datasets

for scenario in scenarios:
    proj_images = np.load(f"../input/preprocessed_2d_proj{scenario}_data_allssp.npy")
    proj_time = pd.read_csv("../input/dates_proj_data.csv")

    proj_season_images, proj_season_time = season_split(
        proj_images, proj_time, "proj", n_memb, scenario
    )



"""
Tests for CERN use case (3DGAN).
"""
import pytest
import subprocess
import os

CERN_PATH = "use-cases/3dgan"
CKPT_NAME = "3dgan-inference.pth"
DEFAULT_DATASET_PATH = 'exp_data'


@pytest.mark.skip("deprecated")
def test_structure_3dgan(check_folder_structure):
    """Test 3DGAN folder structure."""
    check_folder_structure(CERN_PATH)


@pytest.mark.functional
def test_3dgan_train(torch_env, tmp_test_dir, install_requirements):
    """
    Test 3DGAN torch lightning trainer by running it end-to-end.

    If CERN_DATASET env variable is defined, it is used to
    override the default dataset download location: useful
    when it contains a local copy of the dataset, preventing
    downloading it again.
    """
    install_requirements(CERN_PATH, torch_env)
    if os.environ.get('CERN_DATASET'):
        dataset_path = os.environ.get('CERN_DATASET')
    else:
        dataset_path = DEFAULT_DATASET_PATH
    conf = os.path.join(os.path.abspath(CERN_PATH), 'config.yaml')
    cmd = (f"{torch_env}/bin/itwinai exec-pipeline "
           f"--config {conf} --pipe-key training_pipeline "
           f'-o dataset_location={dataset_path} '
           '-o hw_accelerators=auto '
           '-o distributed_strategy=auto '
           '-o mlflow_tracking_uri=ml_logs/mlflow'
           )
    subprocess.run(cmd.split(), check=True, cwd=tmp_test_dir)


@pytest.mark.functional
def test_3dgan_inference(
    torch_env,
    tmp_test_dir,
    install_requirements,
    # fake_model_checkpoint
):
    """
    Test 3DGAN torch lightning trainer by running it end-to-end.

    If CERN_DATASET env variable is defined, it is used to
    override the default dataset download location: useful
    when it contains a local copy of the dataset, preventing
    downloading it again.
    """
    install_requirements(CERN_PATH, torch_env)

    # Create fake inference dataset and checkpoint
    exec = os.path.join(os.path.abspath(CERN_PATH),
                        'create_inference_sample.py')
    cmd = (f"{torch_env}/bin/python {exec} "
           f"--root {tmp_test_dir} "
           f"--ckpt-name {CKPT_NAME}")
    subprocess.run(cmd.split(), check=True, cwd=tmp_test_dir)

    # Test inference
    if os.environ.get('CERN_DATASET'):
        dataset_path = os.environ.get('CERN_DATASET')
    else:
        dataset_path = DEFAULT_DATASET_PATH
    conf = os.path.join(os.path.abspath(CERN_PATH), 'config.yaml')
    cmd = (
        f'{torch_env}/bin/itwinai exec-pipeline '
        f'--config {conf} --pipe-key inference_pipeline '
        f'-o dataset_location={dataset_path} '
        f'-o inference_model_uri={CKPT_NAME} '
        '-o hw_accelerators=auto '
        '-o distributed_strategy=auto '
        '-o logs_dir=ml_logs/mlflow_logs '
        '-o inference_results_location=3dgan-generated-data '
    )
    subprocess.run(cmd.split(), check=True, cwd=tmp_test_dir)



"""
Provide functionalities to manage configuration files, including parsing,
execution, and dynamic override of fields.
"""

import logging
import os
from typing import Dict, Any, List, Type, Union, Optional
from jsonargparse import ArgumentParser as JAPArgumentParser
from jsonargparse import ActionConfigFile
from jsonargparse._formatters import DefaultHelpFormatter

import json
from omegaconf import OmegaConf
from pathlib import Path

from .components import BaseComponent
from .pipeline import Pipeline
from .utils import load_yaml


def add_replace_field(
    config: Dict,
    key_chain: str,
    value: Any
) -> None:
    """Replace or add (if not present) a field in a dictionary, following a
    path of dot-separated keys. Adding is not supported for list items.
    Inplace operation.

    Args:
        config (Dict): dictionary to be modified.
        key_chain (str): path of nested (dot-separated) keys to specify the
            location
            of the new value (e.g., 'foo.bar.line' adds/overwrites the value
            located at config['foo']['bar']['line']).
        value (Any): the value to insert.
    """
    sub_config = config
    for idx, k in enumerate(key_chain.split('.')):
        if idx >= len(key_chain.split('.')) - 1:
            # Last key reached
            break

        if isinstance(sub_config, (list, tuple)):
            k = int(k)
            next_elem = sub_config[k]
        else:
            next_elem = sub_config.get(k)

        if not isinstance(next_elem, (dict, list, tuple)):
            sub_config[k] = dict()

        sub_config = sub_config[k]
    if isinstance(sub_config, (list, tuple)):
        k = int(k)
    sub_config[k] = value


class ConfigParser:
    """
    Parses a pipeline from a configuration file.
    It also provides functionalities for dynamic override
    of fields by means of nested key notation.

    Args:
        config (Union[str, Dict]): path to YAML configuration file
            or dict storing a configuration.
        override_keys (Optional[Dict[str, Any]], optional): dict mapping
            nested keys to the value to override. Defaults to None.

    Example:


    >>> # pipeline.yaml file
    >>> pipeline:
    >>>   class_path: itwinai.pipeline.Pipeline
    >>>   init_args:
    >>>     steps:
    >>>       - class_path: dataloader.MNISTDataModuleTorch
    >>>         init_args:
    >>>           save_path: .tmp/
    >>>
    >>>       - class_path: itwinai.torch.trainer.TorchTrainer
    >>>         init_args:
    >>>           model:
    >>>             class_path: model.Net
    >>>
    >>> from itwinai.parser import ConfigParser
    >>>
    >>> parser = ConfigParser(
    >>>     config='pipeline.yaml',
    >>>     override_keys={
    >>>         'pipeline.init_args.steps.0.init_args.save_path': /save/path
    >>>     }
    >>> )
    >>> pipeline = parser.parse_pipeline()
    >>> print(pipeline)
    >>> print(pipeline.steps)
    >>>
    >>> dataloader = parser.parse_step(0)
    >>> print(dataloader)
    >>> print(dataloader.save_path)

    """

    #: Configuration to parse.
    config: Dict
    #: Pipeline object instantiated from the configuration file.
    pipeline: Pipeline

    def __init__(
        self,
        config: Union[str, Dict],
        override_keys: Optional[Dict[str, Any]] = None
    ) -> None:
        self.config = config
        self.override_keys = override_keys
        if isinstance(self.config, (str, Path)):
            self.config = load_yaml(self.config)
        self._dynamic_override_keys()
        self._omegaconf_interpolate()

    def _dynamic_override_keys(self):
        if self.override_keys is not None:
            for key_chain, value in self.override_keys.items():
                add_replace_field(self.config, key_chain, value)

    def _omegaconf_interpolate(self) -> None:
        """Performs variable interpolation with OmegaConf on internal
        configuration file.
        """
        conf = OmegaConf.create(self.config)
        self.config = OmegaConf.to_container(conf, resolve=True)

    def parse_pipeline(
        self,
        pipeline_nested_key: str = "pipeline",
        verbose: bool = False
    ) -> Pipeline:
        """Merges steps into pipeline and parses it.

        Args:
            pipeline_nested_key (str, optional): nested key in the
                configuration file identifying the pipeline object.
                Defaults to "pipeline".
            verbose (bool): if True, prints the assembled pipeline
                to console formatted as JSON.

        Returns:
            Pipeline: instantiated pipeline.
        """
        pipe_parser = JAPArgumentParser()
        pipe_parser.add_subclass_arguments(Pipeline, "pipeline")

        pipe_dict = self.config
        for key in pipeline_nested_key.split('.'):
            pipe_dict = pipe_dict[key]
        # pipe_dict = self.config[pipeline_nested_key]
        pipe_dict = {"pipeline": pipe_dict}

        if verbose:
            print("Assembled pipeline:")
            print(json.dumps(pipe_dict, indent=4))

        # Parse pipeline dict once merged with steps
        conf = pipe_parser.parse_object(pipe_dict)
        pipe = pipe_parser.instantiate_classes(conf)
        self.pipeline = pipe["pipeline"]
        return self.pipeline

    def parse_step(
        self,
        step_idx: Union[str, int],
        pipeline_nested_key: str = "pipeline",
        verbose: bool = False
    ) -> BaseComponent:
        pipeline_dict = self.config
        for key in pipeline_nested_key.split('.'):
            pipeline_dict = pipeline_dict[key]

        step_dict_config = pipeline_dict['init_args']['steps'][step_idx]

        if verbose:
            print(f"STEP '{step_idx}' CONFIG:")
            print(json.dumps(step_dict_config, indent=4))

        # Wrap config under "step" field and parse it
        step_dict_config = {'step': step_dict_config}
        step_parser = JAPArgumentParser()
        step_parser.add_subclass_arguments(BaseComponent, "step")
        parsed_namespace = step_parser.parse_object(step_dict_config)
        return step_parser.instantiate_classes(parsed_namespace)["step"]


class ArgumentParser(JAPArgumentParser):
    """Wrapper of ``jsonargparse.ArgumentParser``.
    Initializer for ArgumentParser instance. It can parse arguments from
    a series of configuration files. Example:

    >>> python main.py --config base-conf.yaml --config other-conf.yaml \\
    >>> --param OVERRIDE_VAL

    All the arguments from the initializer of `argparse.ArgumentParser
    <https://docs.python.org/3/library/argparse.html#argparse.ArgumentParser>`_
    are supported. Additionally it accepts:

    Args:
        env_prefix (Union[bool, str], optional): Prefix for environment
            variables. ``True`` to derive from ``prog``.. Defaults to True.
        formatter_class (Type[DefaultHelpFormatter], optional): Class for
            printing help messages. Defaults to DefaultHelpFormatter.
        exit_on_error (bool, optional): Defaults to True.
        logger (Union[bool, str, dict, logging.Logger], optional):
            Configures the logger, see :class:`.LoggerProperty`.
            Defaults to False.
        version (Optional[str], optional): Program version which will be
            printed by the --version argument. Defaults to None.
        print_config (Optional[str], optional): Add this as argument to
            print config, set None to disable.
            Defaults to "--print_config".
        parser_mode (str, optional): Mode for parsing config files:
            ``'yaml'``, ``'jsonnet'`` or ones added via
            :func:`.set_loader`.. Defaults to "yaml".
        dump_header (Optional[List[str]], optional): Header to include
            as comment when dumping a config object. Defaults to None.
        default_config_files
            (Optional[List[Union[str, os.PathLike]]], optional):
            Default config file locations, e.g.
            :code:`['~/.config/myapp/*.yaml']`. Defaults to None.
        default_env (bool, optional): Set the default value on whether
            to parse environment variables. Defaults to False.
        default_meta (bool, optional): Set the default value on whether
            to include metadata in config objects. Defaults to True.
    """

    def __init__(
        self,
        *args,
        env_prefix: Union[bool, str] = True,
        formatter_class: Type[DefaultHelpFormatter] = DefaultHelpFormatter,
        exit_on_error: bool = True,
        logger: Union[bool, str, dict, logging.Logger] = False,
        version: Optional[str] = None,
        print_config: Optional[str] = "--print_config",
        parser_mode: str = "yaml",
        dump_header: Optional[List[str]] = None,
        default_config_files: Optional[List[Union[str, os.PathLike]]] = None,
        default_env: bool = False,
        default_meta: bool = True,
        **kwargs,
    ) -> None:
        super().__init__(
            *args, env_prefix=env_prefix, formatter_class=formatter_class,
            exit_on_error=exit_on_error, logger=logger, version=version,
            print_config=print_config, parser_mode=parser_mode,
            dump_header=dump_header, default_config_files=default_config_files,
            default_env=default_env,
            default_meta=default_meta, **kwargs)
        self.add_argument(
            "-c", "--config", action=ActionConfigFile,
            help="Path to a configuration file in json or yaml format."
        )

# type: ignore



"""
Train file to launch pipeline
"""
import os
import sys
from typing import Dict
import argparse
import logging
from datetime import datetime

sys.path.append(os.path.join(os.path.dirname(__file__), 'src'))
sys.path.append(os.path.join(os.path.dirname(__file__), 'preprocessing'))

from itwinai.parser import ConfigParser, ArgumentParser

if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument(
         "-p", "--pipeline", type=str, required=True,
         help='Configuration file to the pipeline to execute.'
    )
    args = parser.parse_args()

    pipe_parser = ConfigParser(
         config=args.pipeline,
    )

    pipeline = pipe_parser.parse_pipeline()
    pipeline.execute()




from typing import Optional
import tensorflow as tf

from .functions import (
    get_tensor_decoding_fn,
    get_scaling_fn,
    get_masking_fn, get_scale_target_fn
)
from ..macros import PatchType, AugmentationType


def get_interleave(cyc_weights, nocyc_weights):
    """
    Returns the interleaved dataset indexes based on cyclone and
    nocyclone weights.
    """
    # define cyclone interleave
    cyc_interleave = [i for i, w in enumerate(cyc_weights) for _ in range(w)]
    # define nocyclone interleave
    nocyc_interleave = [i+len(cyc_interleave)
                        for i, w in enumerate(nocyc_weights) for _ in range(w)]

    # compute the number of blocks + the remainder of the interleaves
    blocks = len(nocyc_interleave) // len(cyc_interleave)
    remainder = len(nocyc_interleave) % len(cyc_interleave)

    interleave = []
    for i in cyc_interleave:
        interleave += [i] + nocyc_interleave[i*blocks:(i+1)*blocks]
    if remainder:
        interleave += nocyc_interleave[-remainder:]

    return tf.cast(interleave, dtype=tf.int64)


def eFlowsTFRecordDataset(
    cyc_fnames, adj_fnames, rnd_fnames,  epochs,  # batch_size,
    scalers, target_scale=False, drv_vars=[], coo_vars=None,
    msk_var=None, shape=(40, 40),
    label_no_cyclone: Optional[float] = -0.3,
    shuffle_buffer=None, patch_type=PatchType.NEAREST.value,
    aug_type=AugmentationType.ONLY_TCS.value, aug_fns={},
    # drop_remainder=True
):
    # set autotune parameter to automatically manage resourches
    AUTOTUNE = tf.data.AUTOTUNE

    # compute the weight associated to an adjacent dataset
    adj_w = 3 if patch_type == PatchType.NEAREST.value else 8

    # setup dynamical lambda functions to be applied to this dataset
    tensor_decoding_fn = get_tensor_decoding_fn(
        shape=shape, drv_vars=drv_vars, coo_vars=coo_vars, msk_var=msk_var)
    scaling_fn = get_scaling_fn(scalers=scalers)
    masking_fn = get_masking_fn(mask=label_no_cyclone)
    scale_target_fn = get_scale_target_fn(
        label_no_cyclone=label_no_cyclone, patch_size=shape[0])

    # multiplier for the augmentation
    mul = 1 if not aug_fns else (len(aug_fns.keys()) + 1)

    # compute the number of samples into the dataset
    cyc_n_elems = sum([int(fname.split('/')[-1].split('.tfrecord')
                      [0].split('_')[-1]) for fname in cyc_fnames])
    rnd_n_elems = sum([int(fname.split('/')[-1].split('.tfrecord')
                      [0].split('_')[-1]) for fname in rnd_fnames])
    adj_n_elems = sum([int(fname.split('/')[-1].split('.tfrecord')
                      [0].split('_')[-1]) for fname in adj_fnames])
    n_elems = mul * cyc_n_elems + rnd_n_elems + adj_n_elems

    # total number of samples that will be yielded by this dataset
    count = n_elems * epochs

    # create standard datasets for each patch category
    cyc_dataset = tf.data.TFRecordDataset(
        cyc_fnames, num_parallel_reads=AUTOTUNE
    ).map(
        tensor_decoding_fn, num_parallel_calls=AUTOTUNE)
    rnd_dataset = tf.data.TFRecordDataset(
        rnd_fnames, num_parallel_reads=AUTOTUNE).map(
        tensor_decoding_fn, num_parallel_calls=AUTOTUNE)
    adj_dataset = tf.data.TFRecordDataset(
        adj_fnames, num_parallel_reads=AUTOTUNE).map(
        tensor_decoding_fn, num_parallel_calls=AUTOTUNE)

    # add cyclone dataset to the cyclone datasets
    cyc_datasets = [cyc_dataset]
    # add the weight of each cyclone dataset
    cyc_weights = [1]

    # add cyclone dataset to the nocyclone datasets
    nocyc_datasets = [rnd_dataset, adj_dataset]
    # add the weight of each nocyclone dataset
    nocyc_weights = [1, adj_w]

    # Create augmented TC datasets and interleave them to the original
    if aug_fns:
        #  augmentation of all patches
        if aug_type == AugmentationType.ALL_PATCHES.value:
            # define augmented datasets for each augmentation function
            aug_cyc_datasets = [
                (tf.data.TFRecordDataset(
                    cyc_fnames, num_parallel_reads=AUTOTUNE).map(
                        tensor_decoding_fn, num_parallel_calls=AUTOTUNE).map(
                    lambda x, y: (aug_fn((x, y))), num_parallel_calls=AUTOTUNE)
                 )
                for aug_fn in aug_fns.values()]

            aug_rnd_datasets = [(tf.data.TFRecordDataset(
                rnd_fnames, num_parallel_reads=AUTOTUNE).map(
                    tensor_decoding_fn, num_parallel_calls=AUTOTUNE).map(
                lambda x, y: (aug_fn((x, y))), num_parallel_calls=AUTOTUNE))
                for aug_fn in aug_fns.values()]

            aug_adj_datasets = [(tf.data.TFRecordDataset(
                adj_fnames, num_parallel_reads=AUTOTUNE).map(
                    tensor_decoding_fn, num_parallel_calls=AUTOTUNE).map(
                lambda x, y: (aug_fn((x, y))), num_parallel_calls=AUTOTUNE))
                for aug_fn in aug_fns.values()]

        # augmentation of only TC patches
        elif aug_type == AugmentationType.ONLY_TCS.value:
            # define augmented datasets for each augmentation function
            aug_cyc_datasets = [(tf.data.TFRecordDataset(
                cyc_fnames, num_parallel_reads=AUTOTUNE).map(
                    tensor_decoding_fn,
                    num_parallel_calls=AUTOTUNE).map(
                lambda x, y: (aug_fn((x, y))),
                num_parallel_calls=AUTOTUNE))
                for aug_fn in aug_fns.values()]
            aug_rnd_datasets = []
            aug_adj_datasets = []
    else:
        aug_cyc_datasets = []
        aug_rnd_datasets = []
        aug_adj_datasets = []

    # add weights from the augmented datasets
    cyc_weights += [1 for _ in range(len(aug_cyc_datasets))]
    nocyc_weights += [1 for _ in range(len(aug_rnd_datasets))]
    nocyc_weights += [adj_w for _ in range(len(aug_adj_datasets))]

    # add to cyclone and nocyclone datasets the augmentations
    cyc_datasets += aug_cyc_datasets
    nocyc_datasets += aug_rnd_datasets + aug_adj_datasets

    # create a list of all datasets to interleave on
    datasets = cyc_datasets + nocyc_datasets

    # get the interleave of all datasets
    interleave = get_interleave(
        cyc_weights=cyc_weights, nocyc_weights=nocyc_weights)

    # compute the choice dataset with the interleave
    choice_dataset = tf.data.Dataset.from_tensor_slices(
        interleave).repeat(count=count)

    # statically interleave elements from all the datasets
    dataset = tf.data.experimental.choose_from_datasets(
        datasets=datasets, choice_dataset=choice_dataset)

    # shuffle if necessary
    if shuffle_buffer:
        dataset = dataset.shuffle(
            shuffle_buffer, reshuffle_each_iteration=True)

    # NOTE: when running distributed training, the dataset
    # should be batched knowing the number of parallel
    # workers taking part to the pool! Skipping it now...
    # Example:
    #  batch_size = num_workers * worker_batch_size
    # # separate in batches
    # if batch_size:
    #     dataset = dataset.batch(
    #         batch_size, drop_remainder=drop_remainder,
    #         num_parallel_calls=AUTOTUNE)
    # else:
    #     dataset = dataset.batch(
    #         n_elems, drop_remainder=drop_remainder,
    #         num_parallel_calls=AUTOTUNE)

    # apply mask on target if label_no_cyclone is provided
    if label_no_cyclone:
        dataset = dataset.map(lambda X, y: (
            masking_fn((X, y))), num_parallel_calls=AUTOTUNE)

    # scale the data
    if scalers:
        dataset = dataset.map(lambda X, y: (
            scaling_fn((X, y))), num_parallel_calls=AUTOTUNE)
    if target_scale:
        dataset = dataset.map(lambda X, y: (
            scale_target_fn((X, y))), num_parallel_calls=AUTOTUNE)

    # set number of epochs that can be repeated on this dataset
    dataset = dataset.repeat(count=epochs)

    # add parallelism option
    options = tf.data.Options()
    options.experimental_distribute.auto_shard_policy = (
        tf.data.experimental.AutoShardPolicy.OFF
    )
    options.experimental_threading.max_intra_op_parallelism = 1
    dataset = dataset.with_options(options)

    # prefetch
    dataset = dataset.prefetch(buffer_size=AUTOTUNE)

    # return the dataset
    return dataset, n_elems



"""Default configuration"""

from pydantic import BaseModel


class Configuration(BaseModel, extra='allow'):
    """Base configuration class."""

    def __getitem__(self, idx):
        return self.__getattribute__(idx)


class TrainingConfiguration(Configuration):
    """Default configuration object for training.
    Override and/or create new configurations using the constructor.

    Example:

    >>> cfg = TrainingConfiguration(batch_size=2, param_a=42)
    >>> print(cfg.batch_size)  # returns 17 (overrides default)
    >>> print(cfg.param_a)     # returns 42 (new value)
    >>> print(cfg.pin_memory)  # returns the default value
    >>>
    >>> from rich import print
    >>> print(cfg)             # pretty-print of configuration

    """
    #: Batch size. In a distributed environment it is usually the
    #: per-worker batch size. Defaults to 32.
    batch_size: int = 32
    #: Whether to shuffle train dataset when creating a torch ``DataLoader``.
    #: Defaults to False.
    shuffle_train: bool = False
    #: Whether to shuffle validation dataset when creating a torch
    #: ``DataLoader``.
    #: Defaults to False.
    shuffle_validation: bool = False
    #: Whether to shuffle test dataset when creating a torch ``DataLoader``.
    #: Defaults to False.
    shuffle_test: bool = False
    #: Whether to pin GPU memory. Property of torch ``DataLoader``.
    #: Defaults to False.
    pin_memory: bool = False
    #: Number of parallel workers used by torch ``DataLoader``.
    #: Defaults to 4.
    num_workers: int = 4
    #: Learning rate used by the optimizer. Defaults to 1e-3.
    lr: float = 1e-3
    #: Momentum used by some optimizers (e.g., SGD). Defaults to 0.9.
    momentum: float = .9
    #: Parameter of Horovod's ``DistributedOptimizer``: uses float16
    #: operations in the allreduce
    #: distributed gradients aggregation. Better performances at
    #: lower precision. Defaults to False.
    fp16_allreduce: bool = False
    #: Parameter of Horovod's ``DistributedOptimizer``: use Adasum
    #: optimization.
    #: Defaults to False.
    use_adasum: bool = False
    #: Parameter of Horovod's ``DistributedOptimizer``: scale
    #: gradients before adding them up.
    #: Defaults to 1.0.
    gradient_predivide_factor: float = 1.0



"""Provides training logic for PyTorch models via Trainer classes."""

from typing import (
    Optional, Dict, Union, Tuple, List, Any, Literal
)
import os
import sys

import torch
from torch.utils.data import DataLoader, Dataset
from torch.utils.data.distributed import DistributedSampler
import torch.distributed as dist
from torch.nn.parallel import DistributedDataParallel as DDP
import torch.nn as nn
from torch.optim.optimizer import Optimizer

import lightning as L
from lightning.pytorch.cli import LightningCLI

import horovod.torch as hvd

from ..components import Trainer, monitor_exec
from .type import (
    Batch, Loss, LrScheduler, Metric
)
from ..loggers import LogMixin, Logger
from .reproducibility import seed_worker, set_seed
from .distributed import (
    TorchDistributedStrategy,
    TorchDDPStrategy,
    HorovodStrategy,
    DeepSpeedStrategy,
    NonDistributedStrategy,
    distributed_resources_available
)
from ..utils import load_yaml
from .mlflow import (
    init_lightning_mlflow,
    teardown_lightning_mlflow
)
from .config import TrainingConfiguration


class TorchTrainer(Trainer, LogMixin):
    """Trainer class for torch training algorithms.

    Args:
        config (Dict): training configuration containing hyperparameters.
        epochs (int): number of training epochs.
        model (Optional[nn.Module], optional): model to train.
            Defaults to None.
        strategy (Literal['ddp', 'deepspeed', 'horovod'], optional):
            distributed strategy. Defaults to 'ddp'.
        validation_every (Optional[int], optional): run a validation epoch
            every ``validation_every`` epochs. Disabled if None. Defaults to 1.
        test_every (Optional[int], optional): run a test epoch
            every ``test_every`` epochs. Disabled if None. Defaults to None.
        random_seed (Optional[int], optional): set random seed for
            reproducibility. If None, the seed is not set. Defaults to None.
        logger (Optional[Logger], optional): logger for ML tracking.
            Defaults to None.
        log_all_workers (bool, optional): if True, the ``log`` method is
            called on all workers in the distributed context.
            Defaults to False.
        metrics (Optional[Dict[str, Metric]], optional): map of torchmetrics
            metrics. Defaults to None.
        checkpoints_location (str): path to checkpoints directory.
            Defaults to "checkpoints".
        checkpoint_every (Optional[int]): save a checkpoint every
            ``checkpoint_every`` epochs. Disabled if None. Defaults to None.
        name (Optional[str], optional): trainer custom name. Defaults to None.
    """
    # TODO:
    #   - extract BaseTorchTrainer and extend it creating a set of trainer
    #     templates (e.g.. GAN, Classifier, Transformer) allowing scientists
    #     to reuse ML algos.

    _strategy: TorchDistributedStrategy = None

    #: PyTorch ``DataLoader`` for training dataset.
    train_dataloader: DataLoader = None
    #: PyTorch ``DataLoader`` for validation dataset.
    validation_dataloader: DataLoader = None
    #: PyTorch ``DataLoader`` for test dataset.
    test_dataloader: DataLoader = None
    #: PyTorch model to train.
    model: nn.Module = None
    #: Loss criterion.
    loss: Loss = None
    #: Optimizer.
    optimizer: Optimizer = None
    #: Learning rate scheduler.
    lr_scheduler: LrScheduler = None
    #: PyTorch random number generator (PRNG).
    torch_rng: torch.Generator = None
    #: itwinai ``itwinai.Logger``
    logger: Logger = None
    #: Total number training batches used so far, across all epochs.
    train_glob_step: int = 0
    #: Total number validation batches used so far, across all epochs.
    validation_glob_step: int = 0
    #: Total number test batches used so far, across all epochs.
    test_glob_step: int = 0
    #: Dictionary of ``torchmetrics`` metrics, indexed by user-defined names.
    metrics: Dict[str, Metric]

    def __init__(
        self,
        config: Dict,
        epochs: int,
        model: Optional[nn.Module] = None,
        strategy: Literal["ddp", "deepspeed", "horovod"] = 'ddp',
        validation_every: Optional[int] = 1,
        test_every: Optional[int] = None,
        random_seed: Optional[int] = None,
        logger: Optional[Logger] = None,
        log_all_workers: bool = False,
        metrics: Optional[Dict[str, Metric]] = None,
        checkpoints_location: str = "checkpoints",
        checkpoint_every: Optional[int] = None,
        name: Optional[str] = None
    ) -> None:
        super().__init__(name)
        self.save_parameters(**self.locals2params(locals()))

        # config is mean to store all hyperparameters, which can very from use
        # case to use case
        # and include learning_rate, batch_size....
        self.config = TrainingConfiguration(**config)
        self.epochs = epochs
        self.model = model
        self.strategy = strategy
        self.validation_every = validation_every
        self.test_every = test_every
        self.random_seed = random_seed
        self.logger = logger
        self.log_all_workers = log_all_workers
        self.metrics = metrics if metrics is not None else {}
        self.checkpoints_location = checkpoints_location
        os.makedirs(self.checkpoints_location, exist_ok=True)
        self.checkpoint_every = checkpoint_every

    @property
    def strategy(self) -> TorchDistributedStrategy:
        """Strategy currently in use."""
        return self._strategy

    @strategy.setter
    def strategy(self, strategy: Union[str, TorchDistributedStrategy]) -> None:
        if isinstance(strategy, TorchDistributedStrategy):
            self._strategy = strategy
        else:
            self._strategy = self._detect_strategy(strategy)

    @property
    def device(self) -> str:
        """Current device from distributed strategy."""
        return self.strategy.device()

    def _detect_strategy(self, strategy: str) -> TorchDistributedStrategy:
        if not distributed_resources_available():
            print("WARNING: falling back to non-distributed strategy.")
            dist_str = NonDistributedStrategy()
        elif strategy == 'ddp':
            dist_str = TorchDDPStrategy(backend='nccl')
        elif strategy == 'horovod':
            dist_str = HorovodStrategy()
        elif strategy == 'deepspeed':
            dist_str = DeepSpeedStrategy(backend='nccl')
        else:
            raise NotImplementedError(
                f"Strategy '{strategy}' is not recognized/implemented.")
        return dist_str

    def _init_distributed_strategy(self) -> None:
        if not self.strategy.is_initialized:
            self.strategy.init()

    def create_model_loss_optimizer(self) -> None:
        """
        Instantiate a torch model, loss, optimizer, and LR scheduler using the
        configuration provided in the Trainer constructor.
        Generally a user-define method.
        """
        ###################################
        # Dear user, this is a method you #
        # may be interested to override!  #
        ###################################

        if self.model is None:
            # Model was not passed to the constructor.
            # Create a model here
            raise ValueError(
                "self.model is None! Either pass it to the constructor or "
                "override this method."
            )

        # A simple NLLLoss
        self.loss = nn.functional.nll_loss

        # TODO: improve robustness of getting from config
        self.optimizer = torch.optim.SGD(
            self.model.parameters(),
            lr=self.config.lr,
            momentum=self.config.momentum
        )
        # Create self.lr_scheduler if needed

        # IMPORTANT: model, optimizer, and scheduler need to be distributed

        # First, define strategy-wise optional configurations
        # TODO: improve robustness of getting from config
        if isinstance(self.strategy, DeepSpeedStrategy):
            # Batch size definition is not optional for DeepSpeedStrategy!
            distribute_kwargs = dict(
                config_params=dict(
                    train_micro_batch_size_per_gpu=self.config.batch_size
                )
            )
        elif isinstance(self.strategy, HorovodStrategy):
            distribute_kwargs = dict(
                compression=(
                    hvd.Compression.fp16 if self.config.fp16_allreduce
                    else hvd.Compression.none
                ),
                op=hvd.Adasum if self.config.use_adasum else hvd.Average,
                gradient_predivide_factor=self.config.gradient_predivide_factor
            )
        else:
            distribute_kwargs = {}

        # Distributed model, optimizer, and scheduler
        (
            self.model,
            self.optimizer,
            self.lr_scheduler
        ) = self.strategy.distributed(
            self.model, self.optimizer, self.lr_scheduler, **distribute_kwargs
        )

    def create_dataloaders(
        self,
        train_dataset: Dataset,
        validation_dataset: Optional[Dataset] = None,
        test_dataset: Optional[Dataset] = None
    ) -> None:
        """
        Create train, validation and test dataloaders using the
        configuration provided in the Trainer constructor.
        Generally a user-define method.

        Args:
            train_dataset (Dataset): training dataset object.
            validation_dataset (Optional[Dataset]): validation dataset object.
                Default None.
            test_dataset (Optional[Dataset]): test dataset object.
                Default None.
        """

        ###################################
        # Dear user, this is a method you #
        # may be interested to override!  #
        ###################################

        self.train_dataloader = self.strategy.create_dataloader(
            dataset=train_dataset,
            batch_size=self.config.batch_size,
            num_workers=self.config.num_workers,
            pin_memory=self.config.pin_memory,
            generator=self.torch_rng,
            shuffle=self.config.shuffle_train
        )
        if validation_dataset is not None:
            self.validation_dataloader = self.strategy.create_dataloader(
                dataset=validation_dataset,
                batch_size=self.config.batch_size,
                num_workers=self.config.num_workers,
                pin_memory=self.config.pin_memory,
                generator=self.torch_rng,
                shuffle=self.config.shuffle_validation
            )
        if test_dataset is not None:
            self.test_dataloader = self.strategy.create_dataloader(
                dataset=test_dataset,
                batch_size=self.config.batch_size,
                num_workers=self.config.num_workers,
                pin_memory=self.config.pin_memory,
                generator=self.torch_rng,
                shuffle=self.config.shuffle_test
            )

    def _setup_metrics(self):
        """Move metrics to current device."""
        for m_name, metric in self.metrics.items():
            self.metrics[m_name] = metric.to(self.device)

    @monitor_exec
    def execute(
        self,
        train_dataset: Dataset,
        validation_dataset: Dataset,
        test_dataset: Dataset
    ) -> Tuple[Dataset, Dataset, Dataset, Any]:
        """Prepares distributed environment and data structures
        for the actual training.

        Args:
            train_dataset (Dataset): training dataset.
            validation_dataset (Dataset): validation dataset.
            test_dataset (Dataset): test dataset.

        Returns:
            Tuple[Dataset, Dataset, Dataset, Any]: training dataset,
            validation dataset, test dataset, trained model.
        """
        self.torch_rng = set_seed(self.random_seed)
        self._init_distributed_strategy()
        self._setup_metrics()

        self.create_dataloaders(
            train_dataset=train_dataset,
            validation_dataset=validation_dataset,
            test_dataset=test_dataset
        )
        self.create_model_loss_optimizer()

        if self.strategy.is_main_worker and self.logger:
            self.logger.create_logger_context()

        self.train()

        if self.strategy.is_main_worker and self.logger:
            self.logger.destroy_logger_context()
        self.strategy.clean_up()
        return train_dataset, validation_dataset, test_dataset, self.model

    def _set_epoch_dataloaders(self, epoch: int):
        """
        Sets epoch in the distributed sampler of a dataloader when using it.
        """
        if self.strategy.is_distributed:
            self.train_dataloader.sampler.set_epoch(epoch)
            if self.validation_dataloader is not None:
                self.validation_dataloader.sampler.set_epoch(epoch)
            if self.test_dataloader is not None:
                self.test_dataloader.sampler.set_epoch(epoch)

    def set_epoch(self, epoch: int) -> None:
        """Set current epoch at the beginning of training.

        Args:
            epoch (int): epoch number, from 0 to ``epochs-1``.
        """
        self._set_epoch_dataloaders(epoch)

    def log(
        self,
        item: Union[Any, List[Any]],
        identifier: Union[str, List[str]],
        kind: str = 'metric',
        step: Optional[int] = None,
        batch_idx: Optional[int] = None,
        **kwargs
    ) -> None:
        """Log ``item`` with ``identifier`` name of ``kind`` type at ``step``
        time step.

        Args:
            item (Union[Any, List[Any]]): element to be logged (e.g., metric).
            identifier (Union[str, List[str]]): unique identifier for the
                element to log(e.g., name of a metric).
            kind (str, optional): type of the item to be logged. Must be one
                among the list of self.supported_types. Defaults to 'metric'.
            step (Optional[int], optional): logging step. Defaults to None.
            batch_idx (Optional[int], optional): DataLoader batch counter
                (i.e., batch idx), if available. Defaults to None.
        """
        if self.logger and (
                self.strategy.is_main_worker or self.log_all_workers):
            self.logger.log(
                item=item,
                identifier=identifier,
                kind=kind,
                step=step,
                batch_idx=batch_idx,
                **kwargs
            )

    def save_checkpoint(
            self, name: str, epoch: int, loss: Optional[torch.Tensor] = None
    ) -> None:
        """Save training checkpoint.

        Args:
            name (str): name of the checkpoint.
            epoch (int): current training epoch.
            loss (Optional[torch.Tensor]): current loss (if available).
        """
        state = dict(
            epoch=epoch,
            loss=loss,
            optimizer=self.optimizer.state_dict(),
            model=self.model.state_dict(),
            lr_scheduler=self.lr_scheduler
        )
        ckpt_path = os.path.join(self.checkpoints_location, name)
        torch.save(state, ckpt_path)
        print(f"Saved '{name}' checkpoint at {ckpt_path}")

        # Save checkpoint to logger
        self.log(ckpt_path, name, kind='artifact')

    def load_checkpoint(self, name: str) -> None:
        """Load state from a checkpoint.

        Args:
            name (str): name of the checkpoint to load, assuming it
                is under ``self.checkpoints_location`` location.
        """
        ckpt_path = os.path.join(self.checkpoints_location, name)
        state = torch.load(ckpt_path, map_location=self.device)
        self.model.load_state_dict(state['model'])
        self.optimizer.load_state_dict(state['optimizer'])
        self.lr_scheduler = state['lr_scheduler']

    def train(self):
        """Trains a machine learning model.
        Main training loop/logic.

        Args:
            train_dataset (Dataset): training dataset.
            validation_dataset (Dataset): validation dataset.
            test_dataset (Dataset): test dataset.

        Returns:
            Tuple[Dataset, Dataset, Dataset, Any]: training dataset,
            validation dataset, test dataset, trained model.
        """
        best_loss = float('inf')
        for epoch in range(self.epochs):
            epoch_n = epoch + 1
            self.set_epoch(epoch)
            self.train_epoch()
            if self.validation_every and epoch_n % self.validation_every == 0:
                val_loss = self.validation_epoch()

                # Checkpointing current best model
                worker_val_losses = self.strategy.gather_obj(
                    val_loss, dst_rank=0)
                if self.strategy.global_rank() == 0:
                    avg_loss = torch.mean(
                        torch.stack(worker_val_losses)
                    ).detach().cpu()
                    if avg_loss < best_loss:
                        ckpt_name = "best_model.pth"
                        self.save_checkpoint(
                            name=ckpt_name, epoch=epoch, loss=avg_loss)

            if self.test_every and epoch_n % self.test_every == 0:
                self.test_epoch()

            # Periodic checkpointing
            if (self.strategy.is_main_worker and self.checkpoint_every
                    and epoch_n % self.checkpoint_every == 0):
                ckpt_name = f"epoch_{epoch}.pth"
                self.save_checkpoint(name=ckpt_name, epoch=epoch)

    def compute_metrics(
        self,
        true: Batch,
        pred: Batch,
        logger_step: int,
        batch_idx: Optional[int],
        stage: str = 'train'
    ) -> Dict[str, Any]:
        """Compute and log metrics.

        Args:
            metrics (Dict[str, Metric]): metrics dict. Can be
                ``self.train_metrics`` or ``self.validation_metrics``.
            true (Batch): true values.
            pred (Batch): predicted values.
            logger_step (int): global step to pass to the logger.
            stage (str): 'train', 'validation'...

        Returns:
            Dict[str, Any]: metric values.
        """
        m_values = {}
        for m_name, metric in self.metrics.items():
            # metric = metric.to(self.device)
            m_val = metric(pred, true).detach().cpu().numpy()
            self.log(
                item=m_val,
                identifier=f'{m_name}_{stage}',
                kind='metric',
                step=logger_step,
                batch_idx=batch_idx
            )
            m_values[m_name] = m_val
        return m_values

    def training_step(
        self,
        batch: Batch,
        batch_idx: int
    ) -> Tuple[Loss, Dict[str, Any]]:
        """Perform a single optimization step using a batch sampled from the
        training dataset.

        Args:
            batch (Batch): batch sampled by a dataloader.
            batch_idx (int): batch index in the dataloader.

        Returns:
            Tuple[Loss, Dict[str, Any]]: batch loss and dictionary of metric
            values with the same structure of ``self.metrics``.
        """
        x, y = batch
        x, y = x.to(self.device), y.to(self.device)
        pred_y = self.model(x)
        loss: Loss = self.loss(pred_y, y)
        self.log(
            item=loss.item(),
            identifier='training_loss',
            kind='metric',
            step=self.train_glob_step,
            batch_idx=batch_idx
        )
        metrics: Dict[str, Any] = self.compute_metrics(
            true=y,
            pred=pred_y,
            logger_step=self.train_glob_step,
            batch_idx=batch_idx,
            stage='training'
        )
        return loss, metrics

    def validation_step(
        self,
        batch: Batch,
        batch_idx: int
    ) -> Tuple[Loss, Dict[str, Any]]:
        """Perform a single optimization step using a batch sampled from the
        validation dataset.

        Args:
            batch (Batch): batch sampled by a dataloader.
            batch_idx (int): batch index in the dataloader.

        Returns:
            Tuple[Loss, Dict[str, Any]]: batch loss and dictionary of metric
            values with the same structure of ``self.metrics``.
        """
        x, y = batch
        x, y = x.to(self.device), y.to(self.device)
        with torch.no_grad():
            pred_y = self.model(x)
            loss: Loss = self.loss(pred_y, y)
        self.log(
            item=loss.item(),
            identifier='validation_loss',
            kind='metric',
            step=self.validation_glob_step,
            batch_idx=batch_idx
        )
        metrics: Dict[str, Any] = self.compute_metrics(
            true=y,
            pred=pred_y,
            logger_step=self.validation_glob_step,
            batch_idx=batch_idx,
            stage='validation'
        )
        return loss, metrics

    def train_epoch(self) -> Loss:
        """Perform a complete sweep over the training dataset, completing an
        epoch of training.

        Returns:
            Loss: average training loss for the current epoch.
        """
        self.model.train()
        train_losses = []
        for batch_idx, train_batch in enumerate(self.train_dataloader):
            loss, metrics = self.training_step(
                batch=train_batch,
                batch_idx=batch_idx
            )
            # TODO: merge and log batch metrics and loss into epoch metrics
            self.optimizer.zero_grad()
            loss.backward()
            self.optimizer.step()
            train_losses.append(loss)
            # Important: update counter
            self.train_glob_step += 1

        # Aggregate and log losses
        avg_loss = torch.mean(torch.stack(train_losses)).detach().cpu()
        self.log(
            item=avg_loss.item(),
            identifier='training_loss_epoch',
            kind='metric',
            step=self.train_glob_step,
        )
        return avg_loss

    def validation_epoch(self) -> Loss:
        """Perform a complete sweep over the validation dataset, completing an
        epoch of validation.

        Returns:
            Loss: average validation loss for the current epoch.
        """
        if self.validation_dataloader is not None:
            self.model.eval()
            validation_losses = []
            for batch_idx, val_batch \
                    in enumerate(self.validation_dataloader):
                # TODO: merge and log batch metrics and loss into epoch metrics
                loss, metrics = self.validation_step(
                    batch=val_batch,
                    batch_idx=batch_idx
                )
                validation_losses.append(loss)
                # Important: update counter
                self.validation_glob_step += 1

            # Aggregate and log losses
            avg_loss = torch.mean(
                torch.stack(validation_losses)
            ).detach().cpu()
            self.log(
                item=avg_loss.item(),
                identifier='validation_loss_epoch',
                kind='metric',
                step=self.validation_glob_step,
            )
            return avg_loss

    def test_epoch(self):
        """Test epoch not implemented yet.

        Raises:
            NotImplementedError: not implemented yet.
        """
        raise NotImplementedError()


class TorchLightningTrainer(Trainer):
    """Generic trainer for torch Lightning workflows.

        Args:
            config (Union[Dict, str]): `Lightning configuration`_
                which can be the path to a file or a Python dictionary.
            mlflow_saved_model (str, optional): name of the model created in
                MLFlow. Defaults to 'my_model'.

        .. _Lightning configuration:
            https://pytorch-lightning.readthedocs.io/en/1.6.5/common/lightning_cli.html
        """

    def __init__(
        self,
        config: Union[Dict, str],
        mlflow_saved_model: str = 'my_model'
    ):
        self.save_parameters(**self.locals2params(locals()))
        super().__init__()
        if isinstance(config, str) and os.path.isfile(config):
            # Load from YAML
            config = load_yaml(config)
        self.conf = config
        self.mlflow_saved_model = mlflow_saved_model

    @monitor_exec
    def execute(self) -> Any:
        init_lightning_mlflow(
            self.conf,
            tmp_dir='/tmp',
            registered_model_name=self.mlflow_saved_model
        )
        old_argv = sys.argv
        sys.argv = ['some_script_placeholder.py']
        cli = LightningCLI(
            args=self.conf,
            model_class=L.LightningModule,
            datamodule_class=L.LightningDataModule,
            run=False,
            save_config_kwargs={
                "overwrite": True,
                "config_filename": "pl-training.yml",
            },
            subclass_mode_model=True,
            subclass_mode_data=True,
        )
        sys.argv = old_argv
        cli.trainer.fit(cli.model, datamodule=cli.datamodule)
        teardown_lightning_mlflow()


def preproc_dataloader(dataloader: DataLoader, gwsize, grank):
    """Makes a Dataloader distributed."""
    sampler = DistributedSampler(
        dataloader.dataset,
        num_replicas=gwsize,
        rank=grank,
        shuffle=True
    )
    # Recreate dataloader, with updated sampler
    return DataLoader(
        dataloader.dataset,
        batch_size=dataloader.batch_size,
        sampler=sampler,
        num_workers=dataloader.num_workers,
        collate_fn=dataloader.collate_fn,
        pin_memory=dataloader.pin_memory,
        drop_last=dataloader.drop_last,
        timeout=dataloader.timeout,
        worker_init_fn=seed_worker,  # dataloader.worker_init_fn,
        multiprocessing_context=dataloader.multiprocessing_context,
        generator=dataloader.generator,
        prefetch_factor=dataloader.prefetch_factor,
        persistent_workers=dataloader.persistent_workers,
        pin_memory_device=dataloader.pin_memory_device
    )


def distributed(func):
    """The decorated function must have a standard signature.
    Its first arguments must be:
    model, train_dataloader, validation_dataloader, device (in this order).

    Additional args or kwargs are allowed consistently with the signature
    of the decorated function.
    """
    def dist_train(
            model, train_dataloader, validation_dataloader=None, device='cpu',
            *args, **kwargs
    ):
        if torch.cuda.is_available():
            dist.init_process_group(backend='nccl')

        if torch.cuda.is_available():
            lwsize = torch.cuda.device_count()  # local world size - per node
            gwsize = dist.get_world_size()     # global world size - per run
            grank = dist.get_rank()            # global rank - assign per run
            lrank = dist.get_rank() % lwsize   # local rank - assign per node
        else:
            gwsize = 1
            grank = 0
            lrank = 0

        device = torch.device(
            'cuda' if torch.cuda.is_available() else 'cpu', lrank)
        if torch.cuda.is_available():
            torch.cuda.set_device(lrank)

        model = model.to(device)
        model = DDP(model, device_ids=[device], output_device=device)

        train_dataloader = preproc_dataloader(train_dataloader, gwsize, grank)
        if validation_dataloader is not None:
            validation_dataloader = preproc_dataloader(
                validation_dataloader, gwsize, grank)

        try:
            func(model, train_dataloader, validation_dataloader, device,
                 *args, **kwargs)
        finally:
            if torch.cuda.is_available():
                dist.barrier()
                dist.destroy_process_group()
    return dist_train



import tensorflow as tf


def custom_VGG_V1(patch_size, channels, activation, regularizer):
    model = tf.keras.Sequential()

    model.add(
        tf.keras.layers.Conv2D(
            input_shape=(patch_size, patch_size, channels[0]),
            filters=64,
            kernel_size=(3, 3),
            padding="same",
            activation=activation,
            kernel_regularizer=regularizer,
        )
    )
    model.add(
        tf.keras.layers.Conv2D(
            filters=64,
            kernel_size=(3, 3),
            padding="same",
            activation=activation,
            kernel_regularizer=regularizer,
        )
    )
    model.add(
        tf.keras.layers.Conv2D(
            filters=64,
            kernel_size=(3, 3),
            padding="same",
            activation=activation,
            kernel_regularizer=regularizer,
        )
    )

    model.add(tf.keras.layers.MaxPooling2D(pool_size=(2, 2), strides=(2, 2)))

    model.add(
        tf.keras.layers.Conv2D(
            filters=128,
            kernel_size=(3, 3),
            padding="same",
            activation=activation,
            kernel_regularizer=regularizer,
        )
    )
    model.add(
        tf.keras.layers.Conv2D(
            filters=128,
            kernel_size=(3, 3),
            padding="same",
            activation=activation,
            kernel_regularizer=regularizer,
        )
    )

    model.add(tf.keras.layers.MaxPooling2D(pool_size=(2, 2), strides=(2, 2)))

    model.add(
        tf.keras.layers.Conv2D(
            filters=256,
            kernel_size=(2, 2),
            padding="same",
            activation=activation,
            kernel_regularizer=regularizer,
        )
    )
    model.add(
        tf.keras.layers.Conv2D(
            filters=256,
            kernel_size=(2, 2),
            padding="same",
            activation=activation,
            kernel_regularizer=regularizer,
        )
    )

    model.add(tf.keras.layers.MaxPooling2D(pool_size=(2, 2), strides=(2, 2)))

    model.add(
        tf.keras.layers.Conv2D(
            filters=512,
            kernel_size=(2, 2),
            padding="valid",
            activation=activation,
            kernel_regularizer=regularizer,
        )
    )
    model.add(
        tf.keras.layers.Conv2D(
            filters=512,
            kernel_size=(2, 2),
            padding="valid",
            activation=activation,
            kernel_regularizer=regularizer,
        )
    )
    model.add(
        tf.keras.layers.Conv2D(
            filters=512,
            kernel_size=(2, 2),
            padding="valid",
            activation=activation,
            kernel_regularizer=regularizer,
        )
    )

    model.add(tf.keras.layers.MaxPooling2D(pool_size=(2, 2)))

    model.add(tf.keras.layers.Flatten())
    model.add(
        tf.keras.layers.Dense(
            units=512, activation=activation, kernel_regularizer=regularizer
        )
    )
    model.add(
        tf.keras.layers.Dense(
            units=256, activation=activation, kernel_regularizer=regularizer
        )
    )
    model.add(
        tf.keras.layers.Dense(
            units=128, activation=activation, kernel_regularizer=regularizer
        )
    )
    model.add(
        tf.keras.layers.Dense(
            units=64, activation=activation, kernel_regularizer=regularizer
        )
    )
    model.add(tf.keras.layers.Dense(channels[1]))

    return model


def custom_VGG_V2(patch_size, channels, activation, regularizer):
    model = tf.keras.Sequential()

    model.add(
        tf.keras.layers.Input(shape=(patch_size, patch_size, channels[0]))
    )

    model.add(
        tf.keras.layers.Conv2D(
            filters=32,
            kernel_size=(3, 3),
            padding="same",
            activation=activation,
            kernel_regularizer=regularizer,
        )
    )
    model.add(
        tf.keras.layers.Conv2D(
            filters=32,
            kernel_size=(3, 3),
            padding="same",
            activation=activation,
            kernel_regularizer=regularizer,
        )
    )
    model.add(
        tf.keras.layers.Conv2D(
            filters=32,
            kernel_size=(3, 3),
            padding="same",
            activation=activation,
            kernel_regularizer=regularizer,
        )
    )

    model.add(tf.keras.layers.MaxPooling2D(pool_size=(2, 2), strides=(2, 2)))

    model.add(
        tf.keras.layers.Conv2D(
            filters=64,
            kernel_size=(3, 3),
            padding="same",
            activation=activation,
            kernel_regularizer=regularizer,
        )
    )
    model.add(
        tf.keras.layers.Conv2D(
            filters=64,
            kernel_size=(3, 3),
            padding="same",
            activation=activation,
            kernel_regularizer=regularizer,
        )
    )
    model.add(
        tf.keras.layers.Conv2D(
            filters=64,
            kernel_size=(3, 3),
            padding="same",
            activation=activation,
            kernel_regularizer=regularizer,
        )
    )

    model.add(tf.keras.layers.MaxPooling2D(pool_size=(2, 2), strides=(2, 2)))

    model.add(
        tf.keras.layers.Conv2D(
            filters=128,
            kernel_size=(3, 3),
            padding="same",
            activation=activation,
            kernel_regularizer=regularizer,
        )
    )
    model.add(
        tf.keras.layers.Conv2D(
            filters=128,
            kernel_size=(3, 3),
            padding="same",
            activation=activation,
            kernel_regularizer=regularizer,
        )
    )
    model.add(
        tf.keras.layers.Conv2D(
            filters=128,
            kernel_size=(3, 3),
            padding="same",
            activation=activation,
            kernel_regularizer=regularizer,
        )
    )

    model.add(tf.keras.layers.MaxPooling2D(pool_size=(2, 2), strides=(2, 2)))

    model.add(
        tf.keras.layers.Conv2D(
            filters=256,
            kernel_size=(2, 2),
            padding="same",
            activation=activation,
            kernel_regularizer=regularizer,
        )
    )
    model.add(
        tf.keras.layers.Conv2D(
            filters=256,
            kernel_size=(2, 2),
            padding="same",
            activation=activation,
            kernel_regularizer=regularizer,
        )
    )
    model.add(
        tf.keras.layers.Conv2D(
            filters=256,
            kernel_size=(2, 2),
            padding="same",
            activation=activation,
            kernel_regularizer=regularizer,
        )
    )

    model.add(
        tf.keras.layers.Conv2D(
            filters=512,
            kernel_size=(2, 2),
            padding="valid",
            activation=activation,
            kernel_regularizer=regularizer,
        )
    )
    model.add(
        tf.keras.layers.Conv2D(
            filters=512,
            kernel_size=(2, 2),
            padding="valid",
            activation=activation,
            kernel_regularizer=regularizer,
        )
    )
    model.add(
        tf.keras.layers.Conv2D(
            filters=512,
            kernel_size=(2, 2),
            padding="valid",
            activation=activation,
            kernel_regularizer=regularizer,
        )
    )
    model.add(
        tf.keras.layers.Conv2D(
            filters=512,
            kernel_size=(2, 2),
            padding="valid",
            activation=activation,
            kernel_regularizer=regularizer,
        )
    )

    model.add(tf.keras.layers.Flatten())

    model.add(
        tf.keras.layers.Dense(
            units=1024, activation=activation, kernel_regularizer=regularizer
        )
    )
    model.add(
        tf.keras.layers.Dense(
            units=512, activation=activation, kernel_regularizer=regularizer
        )
    )
    model.add(
        tf.keras.layers.Dense(
            units=256, activation=activation, kernel_regularizer=regularizer
        )
    )
    model.add(
        tf.keras.layers.Dense(
            units=128, activation=activation, kernel_regularizer=regularizer
        )
    )

    model.add(tf.keras.layers.Dense(channels[1]))

    return model


def custom_VGG_V3(patch_size, channels, activation, regularizer):
    model = tf.keras.Sequential()

    model.add(tf.keras.layers.Input(
        shape=(patch_size, patch_size, channels[0])))

    model.add(
        tf.keras.layers.Conv2D(
            filters=32,
            kernel_size=(3, 3),
            padding="same",
            activation=activation,
            kernel_regularizer=regularizer,
        )
    )
    model.add(
        tf.keras.layers.Conv2D(
            filters=32,
            kernel_size=(3, 3),
            padding="same",
            activation=activation,
            kernel_regularizer=regularizer,
        )
    )
    model.add(
        tf.keras.layers.Conv2D(
            filters=32,
            kernel_size=(3, 3),
            padding="same",
            activation=activation,
            kernel_regularizer=regularizer,
        )
    )

    model.add(tf.keras.layers.MaxPooling2D(pool_size=(2, 2), strides=(2, 2)))

    model.add(
        tf.keras.layers.Conv2D(
            filters=64,
            kernel_size=(3, 3),
            padding="same",
            activation=activation,
            kernel_regularizer=regularizer,
        )
    )
    model.add(
        tf.keras.layers.Conv2D(
            filters=64,
            kernel_size=(3, 3),
            padding="same",
            activation=activation,
            kernel_regularizer=regularizer,
        )
    )
    model.add(
        tf.keras.layers.Conv2D(
            filters=64,
            kernel_size=(3, 3),
            padding="same",
            activation=activation,
            kernel_regularizer=regularizer,
        )
    )

    model.add(tf.keras.layers.MaxPooling2D(pool_size=(2, 2), strides=(2, 2)))

    model.add(
        tf.keras.layers.Conv2D(
            filters=128,
            kernel_size=(3, 3),
            padding="same",
            activation=activation,
            kernel_regularizer=regularizer,
        )
    )
    model.add(
        tf.keras.layers.Conv2D(
            filters=128,
            kernel_size=(3, 3),
            padding="same",
            activation=activation,
            kernel_regularizer=regularizer,
        )
    )
    model.add(
        tf.keras.layers.Conv2D(
            filters=128,
            kernel_size=(3, 3),
            padding="same",
            activation=activation,
            kernel_regularizer=regularizer,
        )
    )

    model.add(tf.keras.layers.MaxPooling2D(pool_size=(2, 2), strides=(2, 2)))

    model.add(
        tf.keras.layers.Conv2D(
            filters=256,
            kernel_size=(3, 3),
            padding="same",
            activation=activation,
            kernel_regularizer=regularizer,
        )
    )
    model.add(
        tf.keras.layers.Conv2D(
            filters=256,
            kernel_size=(3, 3),
            padding="same",
            activation=activation,
            kernel_regularizer=regularizer,
        )
    )
    model.add(
        tf.keras.layers.Conv2D(
            filters=256,
            kernel_size=(3, 3),
            padding="same",
            activation=activation,
            kernel_regularizer=regularizer,
        )
    )

    model.add(
        tf.keras.layers.Conv2D(
            filters=512,
            kernel_size=(2, 2),
            padding="valid",
            activation=activation,
            kernel_regularizer=regularizer,
        )
    )
    model.add(
        tf.keras.layers.Conv2D(
            filters=512,
            kernel_size=(2, 2),
            padding="valid",
            activation=activation,
            kernel_regularizer=regularizer,
        )
    )

    model.add(
        tf.keras.layers.Conv2D(
            filters=1024,
            kernel_size=(2, 2),
            padding="valid",
            activation=activation,
            kernel_regularizer=regularizer,
        )
    )
    model.add(
        tf.keras.layers.Conv2D(
            filters=1024,
            kernel_size=(2, 2),
            padding="valid",
            activation=activation,
            kernel_regularizer=regularizer,
        )
    )

    model.add(tf.keras.layers.Flatten())

    model.add(
        tf.keras.layers.Dense(
            units=1024, activation=activation, kernel_regularizer=regularizer
        )
    )
    model.add(
        tf.keras.layers.Dense(
            units=512, activation=activation, kernel_regularizer=regularizer
        )
    )
    model.add(
        tf.keras.layers.Dense(
            units=512, activation=activation, kernel_regularizer=regularizer
        )
    )
    model.add(
        tf.keras.layers.Dense(
            units=256, activation=activation, kernel_regularizer=regularizer
        )
    )

    model.add(tf.keras.layers.Dense(channels[1]))

    return model


"""
def VGG_V4(patch_size, label_no_cyclone, channels, activation, regularizer):
    model = tf.keras.Sequential()

    model.add(tf.keras.layers.Input(shape=(patch_size, patch_size,
    channels[0])))

    model.add(tf.keras.layers.Conv2D(filters=32, kernel_size=(3,3),
    padding="same", activation=activation, kernel_regularizer=regularizer))
    model.add(tf.keras.layers.Conv2D(filters=32, kernel_size=(3,3),
    padding="same", activation=activation, kernel_regularizer=regularizer))
    model.add(tf.keras.layers.Conv2D(filters=32, kernel_size=(3,3),
    padding="same", activation=activation, kernel_regularizer=regularizer))

    model.add(tf.keras.layers.MaxPooling2D(pool_size=(2,2), strides=(2,2)))

    model.add(tf.keras.layers.Conv2D(filters=64, kernel_size=(3,3),
    padding="same", activation=activation, kernel_regularizer=regularizer))
    model.add(tf.keras.layers.Conv2D(filters=64, kernel_size=(3,3),
    padding="same", activation=activation, kernel_regularizer=regularizer))
    model.add(tf.keras.layers.Conv2D(filters=64, kernel_size=(3,3),
    padding="same", activation=activation, kernel_regularizer=regularizer))

    model.add(tf.keras.layers.MaxPooling2D(pool_size=(2,2), strides=(2,2)))

    model.add(tf.keras.layers.Conv2D(filters=128, kernel_size=(3,3),
    padding="same", activation=activation, kernel_regularizer=regularizer))
    model.add(tf.keras.layers.Conv2D(filters=128, kernel_size=(3,3),
    padding="same", activation=activation, kernel_regularizer=regularizer))
    model.add(tf.keras.layers.Conv2D(filters=128, kernel_size=(3,3),
    padding="same", activation=activation, kernel_regularizer=regularizer))

    model.add(tf.keras.layers.MaxPooling2D(pool_size=(2,2), strides=(2,2)))

    model.add(tf.keras.layers.Conv2D(filters=256, kernel_size=(2,2),
    padding="same", activation=activation, kernel_regularizer=regularizer))
    model.add(tf.keras.layers.Conv2D(filters=256, kernel_size=(2,2),
    padding="same", activation=activation, kernel_regularizer=regularizer))
    model.add(tf.keras.layers.Conv2D(filters=256, kernel_size=(2,2),
    padding="same", activation=activation, kernel_regularizer=regularizer))

    model.add(tf.keras.layers.Conv2D(filters=512, kernel_size=(2,2),
    padding="valid", activation=activation, kernel_regularizer=regularizer))
    model.add(tf.keras.layers.Conv2D(filters=512, kernel_size=(2,2),
    padding="valid", activation=activation, kernel_regularizer=regularizer))
    model.add(tf.keras.layers.Conv2D(filters=512, kernel_size=(2,2),
    padding="valid", activation=activation, kernel_regularizer=regularizer))
    model.add(tf.keras.layers.Conv2D(filters=512, kernel_size=(2,2),
    padding="valid", activation=activation, kernel_regularizer=regularizer))

    model.add(tf.keras.layers.Flatten())

    model.add(tf.keras.layers.Dense(units=1024, activation=activation,
    kernel_regularizer=regularizer))
    model.add(tf.keras.layers.Dense(units=512, activation=activation,
    kernel_regularizer=regularizer))
    model.add(tf.keras.layers.Dense(units=256, activation=activation,
    kernel_regularizer=regularizer))
    model.add(tf.keras.layers.Dense(units=128, activation=activation,
    kernel_regularizer=regularizer))

    model.add(tf.keras.layers.Dense(channels[1]))
    model.add(PositionDiscretizationLayer(label_no_cyclone=label_no_cyclone,
    patch_size=patch_size))

    return model
"""


def ModelV5(patch_size, channels, last_activation, kernel_size=3):
    # kernel initializer
    initializer = tf.random_normal_initializer(0.0, 0.02)

    # input layer
    inputs = tf.keras.layers.Input(shape=(patch_size, patch_size,
channels[0]))

    conv_blocks = [
        ConvBlock(
            filters=32,
            initializer=initializer,
            kernel_size=kernel_size,
            strides=2,
            apply_batchnorm=True,
            apply_dropout=False,
            apply_gaussian_noise=True,
        ),
        ConvBlock(
            filters=64,
            initializer=initializer,
            kernel_size=kernel_size,
            strides=2,
            apply_batchnorm=False,
            apply_dropout=False,
            apply_gaussian_noise=False,
        ),
        ConvBlock(
            filters=128,
            initializer=initializer,
            kernel_size=3,
            strides=2,
            apply_batchnorm=False,
            apply_dropout=True,
            apply_gaussian_noise=False,
        ),
        ConvBlock(
            filters=256,
            initializer=initializer,
            kernel_size=3,
            strides=2,
            apply_batchnorm=False,
            apply_dropout=False,
            apply_gaussian_noise=True,
        ),
        ConvBlock(
            filters=512,
            initializer=initializer,
            kernel_size=3,
            strides=2,
            apply_batchnorm=False,
            apply_dropout=False,
            apply_gaussian_noise=False,
        ),
        ConvBlock(
            filters=1024,
            initializer=initializer,
            kernel_size=3,
            strides=2,
            apply_batchnorm=True,
            apply_dropout=True,
            apply_gaussian_noise=False,
        ),
    ]
    x = inputs
    for block in conv_blocks:
        x = block(x)

    x = tf.keras.layers.Flatten()(x)
    x = tf.keras.layers.Dense(
        units=1024, activation="relu", kernel_initializer=initializer
    )(x)
    x = tf.keras.layers.Dense(
        units=512, activation="relu", kernel_initializer=initializer
    )(x)
    x = tf.keras.layers.Dense(
        units=256, activation="relu", kernel_initializer=initializer
    )(x)
    x = tf.keras.layers.Dense(
        units=128, activation="relu", kernel_initializer=initializer
    )(x)

    outputs = tf.keras.layers.Dense(
        channels[1], activation=last_activation,
kernel_initializer=initializer
    )(x)

    return tf.keras.Model(inputs=inputs, outputs=outputs, name="model_V5")


"""
from tensorflow.keras.layers import Conv2D, MaxPooling2D, Flatten, Dense
from tensorflow.keras.models import Sequential
def custom_VGG_V1(patch_size, channels, activation, regularizer):
    model = Sequential()
    #model.add(Conv2D(input_shape=(40,40,len(variables_list)), filters=64,
    kernel_size=(3,3), padding="same", activation=activation,
    kernel_regularizer=regularizer))
    model.add(Conv2D(input_shape=(patch_size,patch_size,channels[0]),
    filters=64, kernel_size=(3,3), padding="same", activation=activation,
    kernel_regularizer=regularizer))
    model.add(Conv2D(filters=64, kernel_size=(3,3), padding="same",
    activation=activation, kernel_regularizer=regularizer))
    model.add(Conv2D(filters=64, kernel_size=(3,3), padding="same",
    activation=activation, kernel_regularizer=regularizer))
    model.add(MaxPooling2D(pool_size=(2,2), strides=(2,2)))
    model.add(Conv2D(filters=128, kernel_size=(3,3), padding="same",
    activation=activation, kernel_regularizer=regularizer))
    model.add(Conv2D(filters=128, kernel_size=(3,3), padding="same",
    activation=activation, kernel_regularizer=regularizer))
    model.add(MaxPooling2D(pool_size=(2,2), strides=(2,2)))
    model.add(Conv2D(filters=256, kernel_size=(2,2), padding="same",
    activation=activation, kernel_regularizer=regularizer))
    model.add(Conv2D(filters=256, kernel_size=(2,2), padding="same",
    activation=activation, kernel_regularizer=regularizer))
    #model.add(Conv2D(filters=256, kernel_size=(2,2), padding="same",
    activation=activation, kernel_regularizer=regularizer))
    model.add(MaxPooling2D(pool_size=(2,2), strides=(2,2)))
    model.add(Conv2D(filters=512, kernel_size=(2,2), padding="valid",
    activation=activation, kernel_regularizer=regularizer))
    model.add(Conv2D(filters=512, kernel_size=(2,2), padding="valid",
    activation=activation, kernel_regularizer=regularizer))
    model.add(Conv2D(filters=512, kernel_size=(2,2), padding="valid",
    activation=activation, kernel_regularizer=regularizer))
    model.add(MaxPooling2D(pool_size=(2,2))) #,strides=(2,2)))
    # model.add(Conv2D(filters=512, kernel_size=(3,3), padding="same",
    activation="relu"))
    # model.add(Conv2D(filters=512, kernel_size=(3,3), padding="same",
    activation="relu"))
    # model.add(Conv2D(filters=512, kernel_size=(3,3), padding="same",
    activation="relu"))
    # model.add(MaxPooling2D(pool_size=(2,2),strides=(2,2)))
    model.add(Flatten())
    model.add(Dense(units=512, activation=activation,
    kernel_regularizer=regularizer))
    model.add(Dense(units=256, activation=activation,
    kernel_regularizer=regularizer))
    model.add(Dense(units=128, activation=activation,
    kernel_regularizer=regularizer))
    model.add(Dense(units=64, activation=activation,
    kernel_regularizer=regularizer))
    model.add(Dense(channels[1]))

    return model
"""



import os
from typing import Callable
import pytest
import subprocess
import random
import string


FNAMES = [
    'pipeline.yaml',
    'startscript',
]


def rnd_string(len: int = 26):
    return ''.join(random.sample(string.ascii_lowercase, len))


@pytest.fixture
def tmp_test_dir():
    root = '/tmp/pytest'
    os.makedirs(root, exist_ok=True)
    test_dir = os.path.join(root, rnd_string())
    while os.path.exists(test_dir):
        test_dir = os.path.join(root, rnd_string())
    os.makedirs(test_dir, exist_ok=True)

    yield test_dir

    # Optional: remove dir here...


@pytest.fixture
def check_folder_structure() -> Callable:
    """
    Verify that the use case folder complies with some predefined
    structure.
    """
    def _check_structure(root: str):
        for fname in FNAMES:
            fpath = os.path.join(root, fname)
            assert os.path.isfile(fpath), f"'{fname}' is missing in '{fpath}'"
    return _check_structure


@pytest.fixture
def install_requirements() -> Callable:
    """Install requirements.txt, if present in root folder."""
    def _install_reqs(root: str, env_prefix: str):
        req_path = os.path.join(root, 'requirements.txt')
        if os.path.isfile(req_path):
            cmd = f"{env_prefix}/bin/pip install -r {req_path}"
            subprocess.run(cmd.split(), check=True)
    return _install_reqs



"""Dataloader for Torch-based MNIST use case."""

from typing import Optional, Tuple, Callable, Any
import os
import shutil

from PIL import Image
from torch.utils.data import Dataset
from torchvision import transforms, datasets

from itwinai.components import DataGetter, monitor_exec


class MNISTDataModuleTorch(DataGetter):
    """Download MNIST dataset for torch."""

    def __init__(self, save_path: str = '.tmp/',) -> None:
        super().__init__()
        self.save_parameters(**self.locals2params(locals()))
        self.save_path = save_path

    @monitor_exec
    def execute(self) -> Tuple[Dataset, Dataset]:
        train_dataset = datasets.MNIST(
            self.save_path, train=True, download=True,
            transform=transforms.Compose([
                transforms.ToTensor(),
                transforms.Normalize((0.1307,), (0.3081,))
            ]))
        validation_dataset = datasets.MNIST(
            self.save_path, train=False, download=True,
            transform=transforms.Compose([
                transforms.ToTensor(),
                transforms.Normalize((0.1307,), (0.3081,))
            ]))
        print("Train and validation datasets loaded.")
        return train_dataset, validation_dataset, None


class InferenceMNIST(Dataset):
    """Loads a set of MNIST images from a folder of JPG files."""

    def __init__(
        self,
        root: str,
        transform: Optional[Callable] = None,
        supported_format: str = '.jpg'
    ) -> None:
        self.root = root
        self.transform = transform
        self.supported_format = supported_format
        self.data = dict()
        self._load()

    def _load(self):
        for img_file in os.listdir(self.root):
            if not img_file.lower().endswith(self.supported_format):
                continue
            filename = os.path.basename(img_file)
            img = Image.open(os.path.join(self.root, img_file))
            self.data[filename] = img

    def __len__(self) -> int:
        return len(self.data)

    def __getitem__(self, index: int) -> Tuple[Any, Any]:
        """
        Args:
            index (int): Index

        Returns:
            tuple: (image_identifier, image) where image_identifier
                is the unique identifier for the image (e.g., filename).
        """
        img_id, img = list(self.data.items())[index]

        if self.transform is not None:
            img = self.transform(img)

        return img_id, img

    @staticmethod
    def generate_jpg_sample(
        root: str,
        max_items: int = 100
    ):
        """Generate a sample dataset of JPG images starting from
            LeCun's test dataset.

        Args:
            root (str): sample path on disk
            max_items (int, optional): max number of images to
                generate. Defaults to 100.
        """
        if os.path.exists(root):
            shutil.rmtree(root)
        os.makedirs(root)

        test_data = datasets.MNIST(root='.tmp', train=False, download=True)
        for idx, (img, _) in enumerate(test_data):
            if idx >= max_items:
                break
            savepath = os.path.join(root, f'digit_{idx}.jpg')
            img.save(savepath)


class MNISTPredictLoader(DataGetter):
    def __init__(self, test_data_path: str) -> None:
        super().__init__()
        self.save_parameters(**self.locals2params(locals()))
        self.test_data_path = test_data_path

    @monitor_exec
    def execute(self) -> Dataset:
        return InferenceMNIST(
            root=self.test_data_path,
            transform=transforms.Compose([
                transforms.ToTensor(),
                transforms.Normalize((0.1307,), (0.3081,))
            ]))



"""
TensorFlow distributed strategies.
"""

from typing import Tuple
import os
import tensorflow as tf
import tensorflow.distribute as dist


def get_strategy() -> Tuple[tf.distribute.Strategy, int]:
    """Strategy for distributed TensorFlow training. It will automatically
    detect if you are running in a multi-node environment, returning the
    correct TensorFlow distributed strategy for data parallel distributed
    training.

    Returns:
        Tuple[tf.distribute.Strategy, int]: strategy and number of
        `parallel workers`_.

        .. _parallel workers:
            https://stackoverflow.com/questions/66005641/why-we-are-using-strategy-num-replicas-in-sync.
    """

    slurm_jobid = os.environ.get('SLURM_JOB_ID')
    slurm_nnodes = int(os.environ.get('SLURM_NNODES', 0))
    if not slurm_jobid or slurm_nnodes < 2:
        # Single-node environment
        print('Not in SLURM env! Assuming that you '
              'are running on a single node')
        mirrored_strategy = dist.MirroredStrategy()
        return mirrored_strategy, mirrored_strategy.num_replicas_in_sync

    # Multi-node environment in SLURM
    cluster_resolver = dist.cluster_resolver.SlurmClusterResolver(
        port_base=12345)
    implementation = dist.experimental.CommunicationImplementation.NCCL
    communication_options = dist.experimental.CommunicationOptions(
        implementation=implementation)

    # declare distribution strategy
    tf_dist_strategy = dist.MultiWorkerMirroredStrategy(
        cluster_resolver=cluster_resolver,
        communication_options=communication_options
    )

    # number of workers
    n_workers = int(os.environ['SLURM_NTASKS'])
    # list of devices per worker
    devices = tf.config.experimental.list_physical_devices('GPU')
    # number of devices per worker
    n_gpus_per_worker = len(devices)
    # total number of GPUs
    n_gpus = n_workers * n_gpus_per_worker

    # get total number of detected GPUs
    print("Number of detected devices: {}".format(
        n_gpus))

    # get total number of workers
    print("Number of devices: {}".format(
        tf_dist_strategy.num_replicas_in_sync))

    return tf_dist_strategy, tf_dist_strategy.num_replicas_in_sync



"""
Test @distributed function decorator. To run this script, use the following
command:

>>> torchrun --nnodes=1 --nproc_per_node=2 --rdzv_id=100 --rdzv_backend=c10d \
        --rdzv_endpoint=localhost:29400 distribtued_decorator.py

"""

import torch
from torch import nn
import torch.nn.functional as F
from torch.utils.data import DataLoader
from torchvision import transforms, datasets
import torch.optim as optim
from torch.optim.lr_scheduler import StepLR

from itwinai.torch.trainer import distributed


class Net(nn.Module):

    def __init__(self):
        super(Net, self).__init__()
        self.conv1 = nn.Conv2d(1, 10, kernel_size=5)
        self.conv2 = nn.Conv2d(10, 20, kernel_size=5)
        self.conv2_drop = nn.Dropout2d()
        self.fc1 = nn.Linear(320, 50)
        self.fc2 = nn.Linear(50, 10)

    def forward(self, x):
        x = F.relu(F.max_pool2d(self.conv1(x), 2))
        x = F.relu(F.max_pool2d(self.conv2_drop(self.conv2(x)), 2))
        x = x.view(-1, 320)
        x = F.relu(self.fc1(x))
        x = F.dropout(x, training=self.training)
        x = self.fc2(x)
        return F.log_softmax(x, dim=0)


def train(model, device, train_loader, optimizer, epoch):
    model.train()
    for batch_idx, (data, target) in enumerate(train_loader):
        data, target = data.to(device), target.to(device)
        optimizer.zero_grad()
        output = model(data)
        loss = F.nll_loss(output, target)
        loss.backward()
        optimizer.step()
        if batch_idx % 100 == 0:
            print('Train Epoch: {} [{}/{} ({:.0f}%)]\tLoss: {:.6f}'.format(
                epoch, batch_idx * len(data), len(train_loader.dataset),
                100. * batch_idx / len(train_loader), loss.item()))


def test(model, device, test_loader):
    model.eval()
    test_loss = 0
    correct = 0
    with torch.no_grad():
        for data, target in test_loader:
            data, target = data.to(device), target.to(device)
            output = model(data)
            # sum up batch loss
            test_loss += F.nll_loss(output, target, reduction='sum').item()
            # get the index of the max log-probability
            pred = output.argmax(dim=1, keepdim=True)
            correct += pred.eq(target.view_as(pred)).sum().item()

    test_loss /= len(test_loader.dataset)

    print(
        '\nTest set: Average loss: {:.4f}, Accuracy: {}/{} ({:.0f}%)\n'.format(
            test_loss, correct, len(test_loader.dataset),
            100. * correct / len(test_loader.dataset)))


@distributed
def train_func(
        model, train_dataloader, validation_dataloader, device,
        optimizer, scheduler, epochs=10
):
    for epoch in range(1, epochs + 1):
        train(model, device, train_dataloader, optimizer, epoch)
        test(model, device, validation_dataloader)
        scheduler.step()


if __name__ == '__main__':

    train_set = datasets.MNIST(
        '.tmp/', train=True, download=True,
        transform=transforms.Compose([
            transforms.ToTensor(),
            transforms.Normalize((0.1307,), (0.3081,))
        ]))
    val_set = datasets.MNIST(
        '.tmp/', train=False, download=True,
        transform=transforms.Compose([
            transforms.ToTensor(),
            transforms.Normalize((0.1307,), (0.3081,))
        ]))
    model = Net()
    train_dataloader = DataLoader(train_set, batch_size=32, pin_memory=True)
    validation_dataloader = DataLoader(val_set, batch_size=32, pin_memory=True)
    optimizer = optim.Adadelta(model.parameters(), lr=1e-3)
    scheduler = StepLR(optimizer, step_size=1, gamma=0.9)

    # Train distributed
    train_func(model, train_dataloader, validation_dataloader, 'cuda',
               optimizer, scheduler=scheduler, epochs=1)



from typing import Optional
import os
from lightning.pytorch.utilities.types import EVAL_DATALOADERS

import numpy as np
import torch
from torch.utils.data import DataLoader, Dataset
import lightning as pl
import glob
import h5py
import gdown

from itwinai.components import DataGetter, monitor_exec


class Lightning3DGANDownloader(DataGetter):
    def __init__(
        self,
        data_path: str,
        data_url: Optional[str] = None,
        name: Optional[str] = None,
    ) -> None:
        self.save_parameters(**self.locals2params(locals()))
        super().__init__(name)
        self.data_path = data_path
        self.data_url = data_url

    @monitor_exec
    def execute(self):
        # Download data
        if not os.path.exists(self.data_path):
            if self.data_url is None:
                print("WARNING! Data URL is None. "
                      "Skipping dataset downloading")

            gdown.download_folder(
                url=self.data_url, quiet=False,
                output=self.data_path,
                # verify=False
            )


class ParticlesDataset(Dataset):
    def __init__(self, datapath: str, max_samples: Optional[int] = None):
        self.datapath = datapath
        self.max_samples = max_samples
        self.data = dict()

        self.fetch_data()

    def __len__(self):
        return len(self.data["X"])

    def __getitem__(self, idx):
        return {"X": self.data["X"][idx], "Y": self.data["Y"][idx],
                "ang": self.data["ang"][idx], "ecal": self.data["ecal"][idx]}

    def fetch_data(self) -> None:

        print("Searching in :", self.datapath)
        files = sorted(glob.glob(os.path.join(
            self.datapath, '**/*.h5'), recursive=True))
        print("Found {} files. ".format(len(files)))
        if len(files) == 0:
            raise RuntimeError(f"No H5 files found at '{self.datapath}'!")

        # concatenated_datasets = []
        # for datafile in files:
        #     f = h5py.File(datafile, 'r')
        #     dataset = self.GetDataAngleParallel(f)
        #     concatenated_datasets.append(dataset)
        #     # Initialize result dictionary
        #     result = {key: [] for key in concatenated_datasets[0].keys()}
        #     for d in concatenated_datasets:
        #         for key in result.keys():
        #             result[key].extend(d[key])
        # return result

        for datafile in files:
            f = h5py.File(datafile, 'r')
            dataset = self.GetDataAngleParallel(f)
            for field, vals_array in dataset.items():
                if self.data.get(field) is not None:
                    # Resize to include the new array
                    new_shape = list(self.data[field].shape)
                    new_shape[0] += len(vals_array)
                    self.data[field].resize(new_shape)
                    self.data[field][-len(vals_array):] = vals_array
                else:
                    self.data[field] = vals_array

            # Stop loading data, if self.max_samples reached
            if (self.max_samples is not None
                    and len(self.data[field]) >= self.max_samples):
                for field, vals_array in self.data.items():
                    self.data[field] = vals_array[:self.max_samples]
                break

    def GetDataAngleParallel(
        self,
        dataset,
        xscale=1,
        xpower=0.85,
        yscale=100,
        angscale=1,
        angtype="theta",
        thresh=1e-4,
        daxis=-1
    ):
        """Preprocess function for the dataset

        Args:
            dataset (str): Dataset file path
            xscale (int, optional): Value to scale the ECAL values.
                Defaults to 1.
            xpower (int, optional): Value to scale the ECAL values,
                exponentially. Defaults to 1.
            yscale (int, optional): Value to scale the energy values.
                Defaults to 100.
            angscale (int, optional): Value to scale the angle values.
                Defaults to 1.
            angtype (str, optional): Which type of angle to use.
                Defaults to "theta".
            thresh (_type_, optional): Maximum value for ECAL values.
                Defaults to 1e-4.
            daxis (int, optional): Axis to expand values. Defaults to -1.

        Returns:
          Dict: Dictionary containning the preprocessed dataset
        """
        X = np.array(dataset.get("ECAL")) * xscale
        Y = np.array(dataset.get("energy")) / yscale
        X[X < thresh] = 0
        X = X.astype(np.float32)
        Y = Y.astype(np.float32)
        ecal = np.sum(X, axis=(1, 2, 3))
        indexes = np.where(ecal > 10.0)
        X = X[indexes]
        Y = Y[indexes]
        if angtype in dataset:
            ang = np.array(dataset.get(angtype))[indexes]
        # else:
        # ang = gan.measPython(X)
        X = np.expand_dims(X, axis=daxis)
        ecal = ecal[indexes]
        ecal = np.expand_dims(ecal, axis=daxis)
        if xpower != 1.0:
            X = np.power(X, xpower)

        Y = np.array([[el] for el in Y])
        ang = np.array([[el] for el in ang])
        ecal = np.array([[el] for el in ecal])

        final_dataset = {"X": X, "Y": Y, "ang": ang, "ecal": ecal}

        return final_dataset


class ParticlesDataModule(pl.LightningDataModule):
    def __init__(
            self,
            datapath: str,
            batch_size: int,
            num_workers: int = 4,
            max_samples: Optional[int] = None
    ) -> None:
        super().__init__()
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.datapath = datapath
        self.max_samples = max_samples

    def setup(self, stage: str = None):
        # make assignments here (val/train/test split)
        # called on every process in DDP

        if stage == 'fit' or stage is None:
            self.dataset = ParticlesDataset(
                self.datapath,
                max_samples=self.max_samples
            )
            dataset_length = len(self.dataset)
            split_point = int(dataset_length * 0.9)
            self.train_dataset, self.val_dataset = \
                torch.utils.data.random_split(
                    self.dataset, [split_point, dataset_length - split_point])

        if stage == 'predict':
            # TODO: inference dataset should be different in that it
            # does not contain images!
            self.predict_dataset = ParticlesDataset(
                self.datapath,
                max_samples=self.max_samples
            )

        # if stage == 'test' or stage is None:
            # self.test_dataset = MyDataset(self.data_dir, train=False)

    def train_dataloader(self):
        return DataLoader(self.train_dataset, num_workers=self.num_workers,
                          batch_size=self.batch_size, drop_last=True)

    def val_dataloader(self):
        return DataLoader(self.val_dataset, num_workers=self.num_workers,
                          batch_size=self.batch_size, drop_last=True)

    def predict_dataloader(self) -> EVAL_DATALOADERS:
        return DataLoader(self.predict_dataset, num_workers=self.num_workers,
                          batch_size=self.batch_size, drop_last=True)

    # def test_dataloader(self):
        # return DataLoader(self.test_dataset, batch_size=self.batch_size)



import imageio
import numpy as np
import torchvision.transforms as transforms
import matplotlib.pyplot as plt
from torchvision.utils import save_image

to_pil_image = transforms.ToPILImage()

def image_to_vid(images):
    # save evolving images along the learning and get the video
    imgs = [np.array(to_pil_image(img)) for img in images]
    imageio.mimsave('outputs/generated_images.gif', imgs)

def save_reconstructed_images(recon_images, epoch, season = ''):
    # save all reconstructed images at each epoch
    save_image(recon_images.cpu(), f"outputs/image_record/{season}output{epoch}.jpg")

def save_ex(recon_ex, epoch, season = ''):
    # save an example of image at a given epoch
    save_image(recon_ex.cpu(), f"outputs/image_record/{season}ex{epoch}.jpg")

def save_loss_plot(train_loss, valid_loss, season = ''):
    # saves the plot of both losses evolutions
    plt.figure(figsize=(10, 7))
    plt.plot(train_loss, color='orange', label='train loss')
    plt.plot(valid_loss, color='red', label='validation loss')
    plt.xlabel('Epochs')
    plt.ylabel('Loss')
    plt.legend()
    plt.savefig(f'outputs/{season}loss.jpg')
    plt.show()



import torch
import torch.nn as nn
import lightning as L

from torch.nn import functional as F


class MNISTModel(L.LightningModule):
    """
    Simple PL model for MNIST.
    Adapted from
    https://lightning.ai/docs/pytorch/stable/notebooks/lightning_examples/mnist-hello-world.html
    """

    def __init__(
        self,
        hidden_size: int = 64,
    ):
        super().__init__()

        # Automatically save constructor args as hyperparameters
        self.save_hyperparameters()

        # Set our init args as class attributes
        self.hidden_size = hidden_size

        # Hardcode some dataset specific attributes
        self.num_classes = 10
        self.dims = (1, 28, 28)
        channels, width, height = self.dims

        # Define PyTorch model
        self.model = nn.Sequential(
            nn.Flatten(),
            nn.Linear(channels * width * height, hidden_size),
            nn.ReLU(),
            nn.Dropout(0.1),
            nn.Linear(hidden_size, hidden_size),
            nn.ReLU(),
            nn.Dropout(0.1),
            nn.Linear(hidden_size, self.num_classes),
        )

    def forward(self, x):
        x = self.model(x)
        return F.log_softmax(x, dim=1)

    def training_step(self, batch, batch_idx):
        x, y = batch
        logits = self(x)
        loss = F.nll_loss(logits, y)

        # Log metrics with autolog
        self.log("train_loss", loss, on_step=True, on_epoch=True)

        return loss

    def validation_step(self, batch, batch_idx):
        x, y = batch
        logits = self(x)
        loss = F.nll_loss(logits, y)
        # preds = torch.argmax(logits, dim=1)
        self.log("val_loss", loss, prog_bar=True, on_step=True, on_epoch=True)

    def test_step(self, batch, batch_idx):
        x, y = batch
        logits = self(x)
        loss = F.nll_loss(logits, y)
        self.log("test_loss", loss)

    def predict_step(self, batch, batch_idx, dataloader_idx=0):
        x, _ = batch
        logits = self(x)
        preds = torch.argmax(logits, dim=1)
        return preds



"""

THIS SCRIPT IS DEPRECATED! Follow the instructions on README.

Simplified training script: data generation + training in one
procedural script. This is an INTERMEDIATE step of integration in itwinai.
"""
import os
import pandas as pd
import torch
import torch.nn as nn
from torch.utils.data import DataLoader
import time
import numpy as np
from sklearn.model_selection import train_test_split
from tqdm import tqdm

from src.dataset import (
    generate_dataset_aux_channels,
    generate_dataset_main_channel,
    generate_cut_image_dataset,
    normalize_
)
from src.model import UNet  # ,Decoder, Decoder_2d_deep, GeneratorResNet
from src.utils import init_weights, calculate_iou_2d

# Global parameters
DATA_ROOT = "tmp_data"
LOAD_DATASET = False
BATCH_SIZE = 20
LR = 0.00005
SAVE_CHECKPOINT = 'choose_your_path.checkpoint_epoch_{}.pth'
N_EPOCHS = 50


def generate_dataset():
    df_aux_ts = generate_dataset_aux_channels(
        1000, 3, duration=16, sample_rate=500,
        num_waves_range=(20, 25), noise_amplitude=0.6
    )
    df_main_ts = generate_dataset_main_channel(
        df_aux_ts, weights=None, noise_amplitude=0.1
    )

    # save datasets
    save_name_main = 'TimeSeries_dataset_synthetic_main.pkl'
    save_name_aux = 'TimeSeries_dataset_synthetic_aux.pkl'
    df_main_ts.to_pickle(os.path.join(DATA_ROOT, save_name_main))
    df_aux_ts.to_pickle(os.path.join(DATA_ROOT, save_name_aux))

    # Transform to images and save to disk
    df_ts = pd.concat([df_main_ts, df_aux_ts], axis=1)
    df = generate_cut_image_dataset(
        df_ts, list(df_ts.columns),
        num_processes=20, square_size=64
    )
    save_name = 'Image_dataset_synthetic_64x64.pkl'
    df.to_pickle(os.path.join(DATA_ROOT, save_name))


def train_decoder(
    num_epochs, generator, criterion, optimizer, dataloader,
    val_loader, accuracy, checkpoint_path, save_best=True
):
    # num_epochs: (int) number of epochs for training
    # generator: (NN.Module) NN model to train
    # criterion: (torch.optim) optimizer to use in training
    # dataloader: (DataLoader) training data
    # val_loader: (Dataloader) validation data
    # accuracy: (function) metric to measure performance of the model
    #   (Note not to be confused with loss)
    # checkpoint_path: (str) full path (including filename in the form
    #   filename_{}.pkl so to insert num_epoch) to save checkpoints at
    # save_best: (bool) if you want to save best performing model

    # uncomment all lines relative to accuracy if you want to measure IOU
    #   between generated and real spectrograms.
    # Note that it significantly slows down the whole process
    # it also might not work as the function has not been fully implemented yet

    loss_plot = []
    val_loss_plot = []
    acc_plot = []
    val_acc_plot = []
    best_val_loss = 5000000
    for epoch in tqdm(range(1, num_epochs+1)):
        st = time.time()
        epoch_loss = []
        epoch_acc = []
        for i, batch in enumerate(dataloader):
            # batch= transform(batch)
            target = batch[:, 0].unsqueeze(1).to(device)
            # print(f'TARGET ON DEVICE: {target.get_device()}')
            target = target.float()
            input = batch[:, 1:].to(device)
            # print(f'INPUT ON DEVICE: {input.get_device()}')

            optimizer.zero_grad()
            generated = generator(input.float())
            # generated=normalize_(generated,1)
            loss = criterion(generated, target)
            loss.backward()
            optimizer.step()
            epoch_loss.append(loss.detach().cpu().numpy())
            # acc=accuracy(generated.detach().cpu().numpy(),target.detach().cpu().numpy(),20)
            # epoch_acc.append(acc)
        val_loss = []
        val_acc = []
        for batch in (val_loader):
            # batch= transform(batch)
            target = batch[:, 0].unsqueeze(1).to(device)
            target = target.float()
            input = batch[:, 1:].to(device)
            with torch.no_grad():
                generated = generator(input.float())
                # generated=normalize_(generated,1)
                loss = criterion(generated, target)
                val_loss.append(loss.detach().cpu().numpy())
                # acc=accuracy(generated.detach().cpu().numpy(),target.detach().cpu().numpy(),20)
                # val_acc.append(acc)
        loss_plot.append(np.mean(epoch_loss))
        val_loss_plot.append(np.mean(val_loss))
        acc_plot.append(np.mean(epoch_acc))
        val_acc_plot.append(np.mean(val_acc))
        # print('epoch: {} loss: {} val loss: {} accuracy: {} val accuracy:
        # {}'.format(epoch,loss_plot[-1],val_loss_plot[-1],acc_plot[-1],val_acc_plot[-1]))
        et = time.time()
        print('epoch: {} loss: {} val loss: {} time:{}s'.format(
            epoch, loss_plot[-1], val_loss_plot[-1], et-st))

        # Save checkpoint every 100 epochs
        if (epoch+1) % 100 == 0:
            # uncomment the following if you want to save checkpoint every 100
            # epochs regardless of the performance of the model
            # checkpoint = {
            #     'epoch': epoch,
            #     'model_state_dict': generator.state_dict(),
            #     'optimizer_state_dict': optimizer.state_dict(),
            #     'loss': loss_plot[-1],
            #     'val_loss': val_loss_plot[-1],
            # }

            # checkpoint_filename = checkpoint_path.format(epoch)
            # torch.save(checkpoint, checkpoint_filename)

            # instead of val_loss and best_val loss we should use accuracy!!!
            if save_best and val_loss_plot[-1] < best_val_loss:
                # create checkpoint
                checkpoint = {
                    'epoch': epoch,
                    'model_state_dict': generator.state_dict(),
                    'optimizer_state_dict': optimizer.state_dict(),
                    'loss': loss_plot[-1],
                    'val_loss': val_loss_plot[-1],
                }

                # save checkpoint only if it is better than the previous ones
                checkpoint_filename = checkpoint_path.format(epoch)
                torch.save(checkpoint, checkpoint_filename)

                # update best model
                best_val_loss = val_loss_plot[-1]
                best_checkpoint_filename = checkpoint_path.format('best')
                torch.save(checkpoint, best_checkpoint_filename)

    return loss_plot, val_loss_plot, acc_plot, val_acc_plot


if __name__ == "__main__":
    if torch.cuda.is_available():
        device = 'cuda'
        print(torch.cuda.get_device_name(torch.cuda.current_device()))
    else:
        device = 'cpu'

    # Generate dataset
    if not LOAD_DATASET:
        generate_dataset()

    # Split data
    file_path = os.path.join(DATA_ROOT, 'Image_dataset_synthetic_64x64.pkl')
    df = pd.read_pickle(file_path)
    # Convert data to torch
    df = df.applymap(lambda x: torch.tensor(x))

    # Divide Image dataset in main and aux channels. Note that df generated in
    # the section Generate Synthetic Dataset will always have the main channel
    # as its first column
    main_channel = list(df.columns)[0]
    aux_channels = list(df.columns)[1:]

    df_aux_all_2d = pd.DataFrame(df[aux_channels])
    df_main_all_2d = pd.DataFrame(df[main_channel])

    X_train_2d, X_test_2d, y_train_2d, y_test_2d = train_test_split(
        df_aux_all_2d, df_main_all_2d, test_size=0.1, random_state=42)

    # TRAINING SET

    # smaller dataset
    signal_data_train_small_2d = torch.stack([torch.stack(
        [y_train_2d[main_channel].iloc[i]]) for i in range(100)])
    # for i in range(y_train.shape[0])
    aux_data_train_small_2d = torch.stack([torch.stack(
        [X_train_2d.iloc[i][0], X_train_2d.iloc[i][1], X_train_2d.iloc[i][2]])
        for i in range(100)])  # for i in range(X_train.shape[0])

    # whole dataset
    signal_data_train_2d = torch.stack([torch.stack(
        [y_train_2d[main_channel].iloc[i]])
        for i in range(y_train_2d.shape[0])])
    aux_data_train_2d = torch.stack([torch.stack(
        [X_train_2d.iloc[i][0], X_train_2d.iloc[i][1], X_train_2d.iloc[i][2]])
        for i in range(X_train_2d.shape[0])])

    # concatenate torch.tensors
    train_data_2d = torch.cat([signal_data_train_2d, aux_data_train_2d], dim=1)
    train_data_small_2d = torch.cat(
        [signal_data_train_small_2d, aux_data_train_small_2d], dim=1)

    print(signal_data_train_2d.shape)
    print(aux_data_train_2d.shape)
    # --------------------------------------------

    # TEST SET

    # smaller dataset
    signal_data_test_small_2d = torch.stack([torch.stack(
        [y_test_2d[main_channel].iloc[i]]) for i in range(100)])
    # for i in range(y_test.shape[0])
    aux_data_test_small_2d = torch.stack([torch.stack(
        [X_test_2d.iloc[i][0], X_test_2d.iloc[i][1], X_test_2d.iloc[i][2]])
        for i in range(100)])  # for i in range(X_test.shape[0])

    # whole dataset
    signal_data_test_2d = torch.stack([torch.stack(
        [y_test_2d[main_channel].iloc[i]]) for i in range(y_test_2d.shape[0])])
    aux_data_test_2d = torch.stack([torch.stack(
        [X_test_2d.iloc[i][0], X_test_2d.iloc[i][1], X_test_2d.iloc[i][2]])
        for i in range(X_test_2d.shape[0])])

    test_data_2d = torch.cat([signal_data_test_2d, aux_data_test_2d], dim=1)
    test_data_small_2d = torch.cat(
        [signal_data_test_small_2d, aux_data_test_small_2d], dim=1)

    print(signal_data_test_2d.shape)
    print(aux_data_test_2d.shape)
    # -----------------------------------------------

    # Normalize
    train_data_2d = normalize_(train_data_2d)
    test_data_2d = normalize_(test_data_2d)

    # Create dataloader objects with preprocessed dataset
    dataloader = DataLoader(
        train_data_2d,
        batch_size=BATCH_SIZE,
        shuffle=True,
    )
    test_dataloader = DataLoader(
        test_data_2d,
        batch_size=BATCH_SIZE,
        shuffle=False,
    )

    # Model to train
    generator_2d = UNet(
        input_channels=3, output_channels=1, norm=False).to(device)
    init_weights(generator_2d, 'normal', scaling=.02)

    # loss function, learning rate, and optimizer
    l2_loss = nn.MSELoss()  # this is l2!!!
    l1_loss = nn.L1Loss()  # this is L1!!!
    loss = l1_loss  # LogCoshLoss()
    G_optimizer = torch.optim.Adam(generator_2d.parameters(), lr=LR)

    # Training
    loss_plot, val_loss_plot, acc_plot, val_acc_plot = train_decoder(
        N_EPOCHS, generator_2d, loss, G_optimizer, dataloader, test_dataloader,
        calculate_iou_2d, SAVE_CHECKPOINT
    )
